""" Module containing utilities for TED calculations """

from __future__ import annotations

import random
from enum import Enum
from operator import itemgetter
from typing import Any, Callable, Dict, List

import numpy as np
from apted import Config as BaseAptedConfig
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem
from scipy.spatial.distance import jaccard as jaccard_dist

StrDict = Dict[str, Any]


class TreeContent(str, Enum):
    """Possibilities for distance calculations on reaction trees"""

    MOLECULES = "molecules"
    REACTIONS = "reactions"
    BOTH = "both"


class AptedConfig(BaseAptedConfig):
    """
    This is a helper class for the tree edit distance
    calculation. It defines how the substitution
    cost is calculated and how to obtain children nodes.

    :param randomize: if True, the children will be shuffled
    :param sort_children: if True, the children will be sorted
    :param dist_func: the distance function used for renaming nodes, Jaccard by default
    """

    def __init__(
        self,
        randomize: bool = False,
        sort_children: bool = False,
        dist_func: Callable[[np.ndarray, np.ndarray], float] = None,
    ) -> None:
        super().__init__()
        self._randomize = randomize
        self._sort_children = sort_children
        self._dist_func = dist_func or jaccard_dist

    def rename(self, node1: StrDict, node2: StrDict) -> float:
        if node1["type"] != node2["type"]:
            return 1

        fp1 = node1["fingerprint"]
        fp2 = node2["fingerprint"]
        return self._dist_func(fp1, fp2)

    def children(self, node: StrDict) -> List[StrDict]:
        if self._sort_children:
            return sorted(node["children"], key=itemgetter("sort_key"))
        if not self._randomize:
            return node["children"]
        children = list(node["children"])
        random.shuffle(children)
        return children


class StandardFingerprintFactory:
    """
    Calculate Morgan fingerprint for molecules, and difference fingerprints for reactions

    :param radius: the radius of the fingerprint
    :param nbits: the fingerprint lengths
    """

    def __init__(self, radius: int = 2, nbits: int = 2048) -> None:
        self._fp_params = (radius, nbits)

    def __call__(self, tree: StrDict, parent: StrDict = None) -> None:
        if tree["type"] == "reaction":
            if parent is None:
                raise ValueError("Must specify parent when making Morgan fingerprints for reaction nodes")
            self._add_rxn_fingerprint(tree, parent)
        else:
            self._add_mol_fingerprints(tree)

    def _add_mol_fingerprints(self, tree: StrDict) -> None:
        if "fingerprint" not in tree:
            mol = Chem.MolFromSmiles(tree["smiles"])
            rd_fp = AllChem.GetMorganFingerprintAsBitVect(mol, *self._fp_params)
            tree["fingerprint"] = np.zeros((1,), dtype=np.int8)
            DataStructs.ConvertToNumpyArray(rd_fp, tree["fingerprint"])
        tree["sort_key"] = "".join(f"{digit}" for digit in tree["fingerprint"])
        if "children" not in tree:
            tree["children"] = []

        for child in tree["children"]:
            for grandchild in child["children"]:
                self._add_mol_fingerprints(grandchild)

    def _add_rxn_fingerprint(self, node: StrDict, parent: StrDict) -> None:
        if "fingerprint" not in node:
            node["fingerprint"] = parent["fingerprint"].copy()
            for reactant in node["children"]:
                node["fingerprint"] -= reactant["fingerprint"]
        node["sort_key"] = "".join(f"{digit}" for digit in node["fingerprint"])

        for child in node["children"]:
            for grandchild in child.get("children", []):
                self._add_rxn_fingerprint(grandchild, child)



"""
Module containing script to combine raw USPTO files

It will:
    * preserve the ReactionSmiles and Year columns
    * create an ID from PatentNumber and ParagraphNum and row index in the original file
"""

import argparse
from pathlib import Path
from typing import Optional, Sequence

import pandas as pd

from rxnutils.data.uspto.uspto_yield import UsptoYieldCuration

DEFAULT_FILENAMES = [
    "1976_Sep2016_USPTOgrants_smiles.rsmi",
    "2001_Sep2016_USPTOapplications_smiles.rsmi",
]


def main(args: Optional[Sequence[str]] = None) -> None:
    """Function for command-line tool"""
    parser = argparse.ArgumentParser("Script to combine USPTO SMILES data from Figshare")
    parser.add_argument("--filenames", nargs="+", default=DEFAULT_FILENAMES, help="the files to combine")
    parser.add_argument("--output", default="uspto_data.csv", help="the output filename")
    parser.add_argument("--folder", default=".", help="folder with downloaded files")
    parser.add_argument(
        "--with_yields",
        action="store_true",
        default=False,
        help="if to add yield columns",
    )
    args = parser.parse_args(args)

    filenames = [Path(args.folder) / filename for filename in args.filenames]
    data = pd.concat([pd.read_csv(filename, sep="\t", dtype={"ParagraphNum": "str"}) for filename in filenames])

    para_num = data["ParagraphNum"].fillna("")
    row_num = data.index.astype(str)
    data["ID"] = data["PatentNumber"] + ";" + para_num + ";" + row_num
    columns = ["ID", "Year", "ReactionSmiles"]
    if args.with_yields:
        columns += ["TextMinedYield", "CalculatedYield"]
    data2 = data[columns]

    print(f"Total number of unique IDs: {len(set(data2['ID']))}")
    print(f"Total number of records: {len(data2)}")

    if args.with_yields:
        print("Curating yields...")
        data2 = UsptoYieldCuration()(data2)

    data2.to_csv(Path(args.folder) / args.output, sep="\t", index=False)


if __name__ == "__main__":
    main()



"""Module for downloading InChI Trust Reaction InChI."""

import logging
import os
import stat
import sys
from zipfile import ZipFile

import requests

PLATFORM2FOLDER = {
    "linux": "linux",
    "win32": "windows",
}

PLATFORM2EXTENSION = {"linux": "", "win32": ".exe"}

CONFIG = {
    "download_folder": ".",
    "download_url": "https://www.inchi-trust.org/wp/download/RInChI/RInChI-V1-00.zip",
}
PATH = os.path.dirname(__file__)


class RInChIError(Exception):
    """Exception raised by RInChI API"""


def main() -> str:
    """Check if Reaction InchI application is present.
    Download it if it's required to do so.

    :return: Path of the folder containing the appropriate
             command line executable based on system type.
    """
    if sys.platform not in PLATFORM2FOLDER:
        raise RInChIError("RInChI software not supported on this platform")

    rinchi_url = CONFIG.get("download_url", "")
    rinchi_fn = rinchi_url.split("/")[-1]
    download_loc = CONFIG.get("download_folder", "")
    download_loc = os.path.join(PATH, download_loc)
    rinchi_fn = os.path.join(download_loc, rinchi_fn)
    if not os.path.exists(rinchi_fn):
        if not os.path.exists(download_loc):
            logging.info(f"Creating: {download_loc}")
            os.makedirs(download_loc)
        logging.info(f"Downloading: {rinchi_url}")
        req = requests.get(rinchi_url)
        logging.debug(f"{req.status_code}")
        logging.debug(f"{req.headers}")
        req.raise_for_status()
        logging.info(f"Creating: {rinchi_fn}")
        with open(rinchi_fn, "wb") as in_fp:
            in_fp.write(req.content)
        logging.info("Download completed...")
        logging.info(f"Unziping: {rinchi_fn}")
        with ZipFile(rinchi_fn, "r") as out_fp:
            bin_path = [x for x in out_fp.namelist() if x.endswith(_exec_folder_ending(os_sep=False) + "/")]
            logging.debug(bin_path)
            out_fp.extractall(download_loc)
        logging.info("Completed...")
        rinchi_cli_path = os.path.join(download_loc, bin_path[0])
        logging.info(f"RInChI CLI: {rinchi_cli_path}")
        if sys.platform == "linux":
            os.chmod(os.path.join(rinchi_cli_path, "rinchi_cmdline"), stat.S_IEXEC)
    else:
        logging.info(f"RInChI exists at: {rinchi_fn}")
        bin_path = [r for r, d, f in os.walk(download_loc) if r.endswith(_exec_folder_ending(os_sep=True))]
        logging.debug(bin_path)
        rinchi_cli_path = bin_path[0]
        logging.info(f"RInChI CLI: {rinchi_cli_path}")
        if sys.platform == "linux":
            os.chmod(
                os.path.join(rinchi_cli_path, f"rinchi_cmdline{PLATFORM2EXTENSION[sys.platform]}"),
                stat.S_IEXEC,
            )

    return rinchi_cli_path


def _exec_folder_ending(os_sep: bool) -> str:
    sep = os.sep if os_sep else "/"
    return sep.join(["bin", "rinchi_cmdline", f"{PLATFORM2FOLDER[sys.platform]}", "x86_64"])


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    main()



"""Module containg routines and interface to run pipelines"""

import argparse
from typing import Any, Dict, Optional, Sequence

import pandas as pd
import yaml

import rxnutils.pipeline.actions.dataframe_mod  # noqa

# imports needed to register all the actions
# pylint: disable=unused-import
import rxnutils.pipeline.actions.reaction_mod  # noqa
import rxnutils.pipeline.actions.reaction_props  # noqa
import rxnutils.pipeline.actions.templates  # noqa
from rxnutils.data.batch_utils import read_csv_batch
from rxnutils.pipeline.base import create_action, global_apply, list_actions


def run_pipeline(
    data: pd.DataFrame,
    pipeline: Dict[str, Any],
    filename: str,
    save_intermediates: bool = True,
) -> pd.DataFrame:
    """
    Run a given pipeline on a dataset

    The actions are applied sequentials as they are defined in the pipeline

    The intermediate results of the pipeline will be written to separate
    tab-separated CSV files.

    :param data: the dataset
    :param pipeline: the action specifications
    :param filename: path to the final output file
    :param save_intermediates: if True will save intermediate results
    :return: the dataset after completing the pipeline
    """
    actions = [create_action(name, **(options or {})) for name, options in pipeline.items()]

    for idx, action in enumerate(actions):
        print(f"Running {action}", flush=True)
        data = action(data)
        if save_intermediates:
            data.to_csv(f"{filename}.{idx}", index=False, sep="\t")
    return data


def main(args: Optional[Sequence[str]] = None) -> None:
    """Function for command line argument"""
    parser = argparse.ArgumentParser("Runner of validation pipeline")
    parser.add_argument("--pipeline", help="the yaml file with a pipeline")
    parser.add_argument("--data", help="the data to be processed. Should be a tab-separated CSV-file")
    parser.add_argument("--output", help="the processed data. Will be a tab-separated CSV-file")
    parser.add_argument("--max-workers", type=int, help="the maximum number of works")
    parser.add_argument(
        "--batch",
        type=int,
        nargs=2,
        help="Line numbers to start and stop reading rows (1-indexed to always include the header, start:end).",
    )
    parser.add_argument("--no-intermediates", action="store_true", default=False)
    parser.add_argument("--list", action="store_true", default=False)
    parser.add_argument("--short-list", action="store_true", default=False)
    args = parser.parse_args(args)

    if args.list or args.short_list:
        print("The available actions are as follow:")
        list_actions(short=args.short_list)
        return

    global_apply.max_workers = args.max_workers

    with open(args.pipeline, "r") as fileobj:
        pipeline = yaml.load(fileobj, Loader=yaml.SafeLoader)

    data = read_csv_batch(args.data, sep="\t", index_col=False, batch=args.batch)

    data = run_pipeline(data, pipeline, args.output, not args.no_intermediates)
    data.to_csv(args.output, index=False, sep="\t")


if __name__ == "__main__":
    main()



"""Module containing useful representations of templates
"""

import hashlib
import logging
import re
from collections import defaultdict
from itertools import permutations
from typing import Any, Dict, Iterator, List, Set, Tuple

import numpy as np
import rdchiral.main as rdc
from rdkit import Chem, DataStructs
from rdkit.Chem import AllChem, SanitizeFlags  # pylint: disable=all
from xxhash import xxh32

DELIM_REGEX_STR = r"[&:\]]"
AROMATIC_REGEX_STR = r"&a" + DELIM_REGEX_STR
AROMATIC_REGEX = re.compile(AROMATIC_REGEX_STR)
CHARGE_REGEX_STR = r"&([+-]\d?)" + DELIM_REGEX_STR
CHARGE_REGEX = re.compile(CHARGE_REGEX_STR)
HYDROGEN_REGEX_STR = r"&H(\d)" + DELIM_REGEX_STR
HYDROGEN_REGEX = re.compile(HYDROGEN_REGEX_STR)
DEGREE_REGEX_STR = r"&D(\d)" + DELIM_REGEX_STR
DEGREE_REGEX = re.compile(DEGREE_REGEX_STR)


class TemplateMolecule:
    """
    Representation of a molecule created from a SMARTS string

    :param rd_mol: the RDKit molecule to be represented by this class
    """

    def __init__(self, rd_mol: Chem.rdchem.Mol = None, smarts: str = None) -> None:
        if rd_mol is None and smarts is None:
            raise ValueError("Both rd_mol and smarts argument is None")

        if rd_mol is None:
            self._rd_mol = Chem.MolFromSmarts(smarts)
            self._smarts = smarts
        else:
            self._rd_mol = rd_mol
            self._smarts = Chem.MolToSmarts(self._rd_mol)

    def atoms(self) -> Iterator[Chem.rdchem.Atom]:
        """
        Generate the atom object of this molecule

        :yield: the next atom object
        """
        for atom in self._rd_mol.GetAtoms():
            yield atom

    def atom_invariants(self) -> List[int]:
        """
        Calculate invariants on similar properties as in RDKit but ignore mass and add aromaticity

        :return: a list of the atom invariants
        """
        self.fix_atom_properties()
        invariants = []
        for atom in self.atoms():
            arr = np.asarray(
                [
                    atom.GetAtomicNum(),
                    atom.GetIntProp("comp_degree"),
                    atom.GetTotalNumHs(),
                    atom.GetFormalCharge(),
                    int(atom.IsInRing()),
                    int(atom.GetIsAromatic()),
                ],
                dtype=np.uint8,
            )
            invariants.append(xxh32(arr.tobytes()).intdigest())
        return invariants

    def atom_properties(self) -> Dict[str, List[object]]:
        """
        Return a dictionary with atomic properties

        Example:
            import pandas
            pandas.DataFrame(my_mol.atom_properties())
        """
        props: Dict[str, List[Any]] = defaultdict(list)
        for idx, atom in enumerate(self.atoms()):
            props["index"].append(idx)
            for key in dir(atom):
                if key.startswith("Get") and "Prop" not in key:
                    ret = getattr(atom, key)()
                    if isinstance(ret, (list, tuple)):
                        props[f"# {key[3:]}"].append(len(ret))
                    else:
                        props[key[3:]].append(ret)
            try:
                props["comp degree"].append(atom.GetIntProp("comp_degree"))
            except KeyError:  # Raised if fix_atom_properties hasn't been called
                pass
        return dict(props)

    def fingerprint_bits(self, radius: int = 2, use_chirality: bool = True) -> Set[int]:
        """
        Calculate the unique fingerprint bits

        Will sanitize molecule if necessary

        :param radius: the radius of the Morgan calculation
        :param use_chirality: determines if chirality should be taken into account
        :return: the set of unique bits
        """

        def calc_fp():
            AllChem.GetMorganFingerprint(
                self._rd_mol,
                radius,
                bitInfo=bits,
                invariants=invariants,
                useChirality=use_chirality,
            )

        invariants = self.atom_invariants()
        bits: Dict[int, List[Tuple[int, int]]] = {}
        try:
            calc_fp()
        except RuntimeError:
            self.sanitize()
            calc_fp()
        return set(bits.keys())

    def fingerprint_vector(self, radius: int = 2, nbits: int = 1024, use_chirality: bool = True) -> np.ndarray:
        """
        Calculate the finger bit vector

        Will sanitize molecule if necessary

        :param radius: the radius of the Morgan calculation
        :param nbits: the length of the bit vector
        :param use_chirality: determines if chirality should be taken into account
        :return: the bit vector
        """

        def calc_fp():
            return AllChem.GetMorganFingerprintAsBitVect(
                self._rd_mol,
                radius,
                nbits,
                invariants=invariants,
                useChirality=use_chirality,
            )

        invariants = self.atom_invariants()
        try:
            bitvect = calc_fp()
        except RuntimeError:
            self.sanitize()
            bitvect = calc_fp()
        array = np.zeros((1,))
        DataStructs.ConvertToNumpyArray(bitvect, array)
        return array

    def fix_atom_properties(self) -> None:
        """
        Copy over some properties from the SMARTS specification to the atom object
        1. Set IsAromatic flag is lower-case a is in the SMARTS
        2. Fix formal charges
        3. Explicit number of hydrogen atoms

        Also extract explicit degree from SMARTS and is stored in
        the `comp_degree` property.
        """
        for atom in self.atoms():
            atom.SetNoImplicit(False)
            aromatic_match = AROMATIC_REGEX.search(atom.GetSmarts())
            if aromatic_match:
                atom.SetIsAromatic(True)

            charge_match = CHARGE_REGEX.search(atom.GetSmarts())
            if charge_match:
                amount = "1" if len(charge_match.group(1)) == 1 else ""
                charge = int(charge_match.group(1) + amount)
                atom.SetFormalCharge(charge)

            hydrogen_match = HYDROGEN_REGEX.search(atom.GetSmarts())
            if hydrogen_match:
                atom.SetNoImplicit(True)
                nhydrogens = int(hydrogen_match.group(1))
                atom.SetNumExplicitHs(nhydrogens)

            degree_match = DEGREE_REGEX.search(atom.GetSmarts())
            if degree_match:
                degrees = int(degree_match.group(1))
                atom.SetIntProp("comp_degree", degrees)
            else:
                atom.SetIntProp("comp_degree", atom.GetDegree())
            atom.UpdatePropertyCache()

    def hash_from_smiles(self) -> str:
        """
        Create a hash of the template based on a cleaned-up template SMILES string

        :return: the hash string
        """
        other_template = TemplateMolecule(smarts=self._smarts)
        other_template.remove_atom_mapping()
        # pylint: disable=protected-access
        smiles = Chem.MolToSmiles(other_template._rd_mol)
        return hashlib.sha224(smiles.encode("utf8")).hexdigest()

    def hash_from_smarts(self) -> str:
        """
        Create a hash of the template based on a cleaned-up template SMARTS string

        :return: the hash string
        """
        other_template = TemplateMolecule(smarts=self._smarts)
        other_template.remove_atom_mapping()
        # pylint: disable=protected-access
        smarts = Chem.MolToSmarts(other_template._rd_mol)
        return hashlib.sha224(smarts.encode("utf8")).hexdigest()

    def remove_atom_mapping(self) -> None:
        """Remove the atom mappings from the molecule"""
        for atom in self.atoms():
            atom.SetAtomMapNum(0)

    def sanitize(self) -> None:
        """
        Will do selective sanitation - skip some procedures that causes problems due to "hanging" aromatic atoms

        All possible flags:
            SANITIZE_ADJUSTHS
            SANITIZE_ALL
            SANITIZE_CLEANUP
            SANITIZE_CLEANUPCHIRALITY
            SANITIZE_FINDRADICALS
            SANITIZE_KEKULIZE
            SANITIZE_NONE
            SANITIZE_PROPERTIES
            SANITIZE_SETAROMATICITY
            SANITIZE_SETCONJUGATION
            SANITIZE_SETHYBRIDIZATION
            SANITIZE_SYMMRINGS
        """
        AllChem.SanitizeMol(
            self._rd_mol,
            sanitizeOps=SanitizeFlags.SANITIZE_ADJUSTHS
            | SanitizeFlags.SANITIZE_CLEANUP
            | SanitizeFlags.SANITIZE_CLEANUPCHIRALITY
            | SanitizeFlags.SANITIZE_PROPERTIES
            | SanitizeFlags.SANITIZE_SETAROMATICITY
            | SanitizeFlags.SANITIZE_SETCONJUGATION
            | SanitizeFlags.SANITIZE_SETHYBRIDIZATION
            | SanitizeFlags.SANITIZE_SYMMRINGS,
        )


class ReactionTemplate:
    """
    Representation of a reaction template created with RDChiral

    :param smarts: the SMARTS string representation of the reaction
    :param direction: if equal to "retro" reverse the meaning of products and reactants
    """

    def __init__(self, smarts: str, direction: str = "canonical") -> None:
        self.smarts = smarts
        self.direction = direction
        self._rd_reaction = AllChem.ReactionFromSmarts(smarts)
        self._reactants = [
            TemplateMolecule(self._rd_reaction.GetReactantTemplate(i))
            for i in range(self._rd_reaction.GetNumReactantTemplates())
        ]
        self._products = [
            TemplateMolecule(self._rd_reaction.GetProductTemplate(i))
            for i in range(self._rd_reaction.GetNumProductTemplates())
        ]

        if direction == "retro":
            temp = self._products
            self._products = self._reactants
            self._reactants = temp

    def apply(self, mols: str) -> Tuple[Tuple[str, ...], ...]:
        """
        Applies the template on the given molecule

        :param mols: the molecule as a SMILES
        :return: the list of reactants
        """
        if self.direction == "retro":
            outcome = rdc.rdchiralRunText(self.smarts, mols)
            return tuple(tuple(str_.split(".")) for str_ in outcome)

        mols_objs = [Chem.MolFromSmiles(mol) for mol in mols.split(".")]
        # Get all permutations of molecules
        outcome = []
        num_reactant_templates = self._rd_reaction.GetNumReactantTemplates()
        logging.debug(f"#Reactants: {len(mols_objs)} Vs. #Reactant Templates: {num_reactant_templates}")
        reactants_permutations = permutations(mols_objs, num_reactant_templates)
        for idx, reactants in enumerate(reactants_permutations):
            outcome = self._rd_reaction.RunReactants(reactants)
            # Break as soon as one outcome has been produced
            # or 100 (0-99) permutations have been tried...
            if outcome or idx == 99:
                break

        def create_smiles(mol_list):
            smiles_list: List[str] = []
            for mol in mol_list:
                try:
                    Chem.SanitizeMol(mol)
                except Exception as error:  # pylint: disable=broad-except
                    logging.error(f"{error}")
                else:
                    smiles_list.append(Chem.MolToSmiles(mol))

            return tuple(smiles_list)

        return tuple(create_smiles(list_) for list_ in outcome)

    def fingerprint_bits(self, radius: int = 2, use_chirality: bool = True) -> Dict[int, int]:
        """
        Calculate the difference count of the fingerprint bits set of the reactants and products

        :param radius: the radius of the Morgan calculation
        :param use_chirality: determines if chirality should be taken into account
        :return: a dictionary of the difference count for each bit
        """
        bit_sum = defaultdict(int)
        for mol in self._products:
            for bit in mol.fingerprint_bits(radius, use_chirality):
                bit_sum[bit] = 1
        for mol in self._reactants:
            for bit in mol.fingerprint_bits(radius, use_chirality):
                bit_sum[bit] -= 1
        return bit_sum

    def fingerprint_vector(self, radius: int = 2, nbits: int = 1024, use_chirality: bool = True) -> np.ndarray:
        """
        Calculate the difference fingerprint vector

        :param radius: the radius of the Morgan calculation
        :param nbits: the length of the bit vector
        :param use_chirality: determines if chirality should be taken into account
        :return: the bit vector
        """
        fp_ = np.zeros(nbits)
        for mol in self._products:
            fp_ += mol.fingerprint_vector(radius, nbits, use_chirality)
        for mol in self._reactants:
            fp_ -= mol.fingerprint_vector(radius, nbits, use_chirality)
        return fp_

    def hash_from_bits(self, radius: int = 2, use_chirality: bool = True) -> str:
        """
        Create a hash of the template based on the difference counts of the fingerprint bits

        :param radius: the radius of the Morgan calculation
        :param use_chirality: determines if chirality should be taken into account
        :return: the hash string
        """
        bit_sum = self.fingerprint_bits(radius, use_chirality)
        hash_obj = hashlib.sha256()
        sorted_bits = sorted(bit_sum.keys())
        for bit in sorted_bits:
            hash_obj.update(f"{bit}:{bit_sum[bit]}".encode())
        return hash_obj.hexdigest()

    def hash_from_smiles(self) -> str:
        """
        Create a hash of the template based on a cleaned-up template SMILES string

        :return: the hash string
        """
        reaction = self._cleaned_reaction_copy()
        AllChem.RemoveMappingNumbersFromReactions(reaction)
        rxn_smiles = AllChem.ReactionToSmiles(reaction)
        return hashlib.sha224(rxn_smiles.encode()).hexdigest()

    def hash_from_smarts(self) -> str:
        """
        Create a hash of the template based on a cleaned-up template SMARTS string

        :return: the hash string
        """
        reaction = self._cleaned_reaction_copy()
        AllChem.RemoveMappingNumbersFromReactions(reaction)
        rxn_smarts = AllChem.ReactionToSmarts(reaction)
        return hashlib.sha224(rxn_smarts.encode()).hexdigest()

    def _cleaned_reaction_copy(self) -> AllChem.ChemicalReaction:
        """Clean RDKit chemical reaction.

        :return: RDKit reaction
        :rtype: AllChem.ChemicalReaction
        """
        reaction = AllChem.ReactionFromSmarts(self.smarts)
        mols = [TemplateMolecule(reaction.GetProductTemplate(i)) for i in range(reaction.GetNumProductTemplates())]
        mols.extend(
            [TemplateMolecule(reaction.GetReactantTemplate(i)) for i in range(reaction.GetNumReactantTemplates())]
        )

        for mol in mols:
            mol.fix_atom_properties()

        return reaction

    def rdkit_validation(self) -> bool:
        """Checks if the template is valid in RDKit"""
        rxn = AllChem.ReactionFromSmarts(self.smarts)
        rxn.Initialize()
        return rxn.Validate()[1] == 0



from __future__ import annotations

import re
from typing import List, Tuple

from rdkit import Chem

from rxnutils.chem.utils import remove_atom_mapping


def smiles_tokens(smiles: str) -> List[str]:
    """
    Tokenize SMILES using basic regex pattern for Chemformer.

    :param smiles: SMILES to tokenize
    :return: List of tokens identified in SMILES.
    """
    pattern = (
        r"(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\\\|\\|\/|:|~|@|\?|>|\*|\!|\$|\%[0-9]{2}|[0-9])"
    )
    regex = re.compile(pattern)
    tokens = [token for token in regex.findall(smiles)]

    tokenized_smiles = "".join(tokens)
    if smiles != tokenized_smiles:
        raise AssertionError(
            f"tokenized SMILES not the same as input SMILES: {tokenized_smiles}, " "{smiles}, tokens: {tokens}"
        )
    return tokens


def _next_tagged_token(product_tagged_tokens: List[str], untagged_token: str, tagged_token_idx: int) -> Tuple[str, int]:
    """
    Get the next tagged token in the sequence. Includes checks and fixes for
    stereochemistry changes due to removing atom mapping.

    :param product_tagged_tokens: tokens of product tagged with [<atom>:1]
    :param untagged_token: the current token from the untagged product
    :param tagged_token_idx: the current token index of the tagged product
    :return: the next (tagged-product) token and the corresponding token index
    """
    tagged_token = product_tagged_tokens[tagged_token_idx]

    # Check if the stereo chemistry has changed after removing atom-mapping and
    # handle each specific case.
    if tagged_token != untagged_token and (tagged_token == "/" or tagged_token == "\\"):
        if untagged_token == "/" or untagged_token == "\\":
            return untagged_token, tagged_token_idx
        else:
            tagged_token_idx += 1
            return product_tagged_tokens[tagged_token_idx], tagged_token_idx

    if tagged_token != untagged_token and not ":1" in tagged_token and "@" in tagged_token:
        return untagged_token, tagged_token_idx

    return tagged_token, tagged_token_idx


def tagged_smiles_from_tokens(product_tagged_tokens: List[str], product_untagged_tokens: List[str]) -> Tuple[str, str]:
    """
    Convert the tagged SMILES from atom-mapping to unmapped-token + '!'

    :param product_tagged_tokens: tokens of product tagged with [<atom>:1]
    :param product_untagged_tokens: tokens of the untagged product

    :return: Tuple of SMILES of the product containing tags corresponding to atoms changed in the
        reaction using "<atom>!", and SMILES of the (reconstructed) untagged product
    """

    product_converted = ""
    product_untagged = ""

    tagged_token_idx = 0

    for untagged_token in product_untagged_tokens:

        tagged_token, tagged_token_idx = _next_tagged_token(product_tagged_tokens, untagged_token, tagged_token_idx)

        if tagged_token != untagged_token and (untagged_token == "/" or untagged_token == "\\"):
            continue

        if tagged_token == untagged_token:
            product_converted += untagged_token
        else:
            # Remove brackets around a single letter
            if len(untagged_token) == 3 and untagged_token.startswith("[") and untagged_token.endswith("]"):
                untagged_token = untagged_token[1]
            product_converted += untagged_token + "!"

        product_untagged += untagged_token

        tagged_token_idx += 1

    return product_converted, product_untagged


def _canonicalize_tagged_smiles(product_tagged: str, product_untagged: str = None) -> Tuple[str, str]:
    """
    Reorder the tagged-product SMILES on canonical form using the canonicalized
    untagged product.

    :param product_tagged: SMILES of tagged product
    :param product_untagged: SMILES of untagged product
    :return: canonicalized untagged and tagged product SMILES
    """
    mol = Chem.MolFromSmiles(product_tagged)
    mol_untagged = Chem.MolFromSmiles(product_untagged)

    _, canonical_atom_order = tuple(zip(*sorted([(j, i) for i, j in enumerate(Chem.CanonicalRankAtoms(mol_untagged))])))

    mol = Chem.RenumberAtoms(mol, canonical_atom_order)
    mol_untagged = Chem.RenumberAtoms(mol_untagged, canonical_atom_order)
    return Chem.MolToSmiles(mol, canonical=False), Chem.MolToSmiles(mol_untagged)


def convert_atom_map_tag(product_atom_map_tagged: str) -> str:
    """
    Replace product tagged by atom-mapping [<atom>:1] to product tagged by "<atom>!".
    Returns empty string if no atom-map tagging or the failed to create untagged product.

    :param product_tagged: SMILES of the product containing tags corresponding to
        atoms changed in the reaction using [<atom>:1]
    :return: SMILES of the product containing tags corresponding to atoms changed in the
        reaction using "<atom>!"
    """

    # Check number of tags
    n_tags = len(re.findall(r"\[[^\]]+:1]", product_atom_map_tagged))

    if n_tags < 1:
        return ""

    product_untagged = remove_atom_mapping(product_atom_map_tagged, canonical=False)

    if not Chem.MolFromSmiles(product_untagged):
        return ""

    product_tagged, product_untagged = _canonicalize_tagged_smiles(product_atom_map_tagged, product_untagged)

    # Update the SMILES string to remove atom-mapping brackets and explicit [H]:s and
    # replace by <atom>!
    product_tagged_tokens = smiles_tokens(product_tagged)
    product_untagged_tokens = smiles_tokens(product_untagged)

    product_tagged_converted, product_untagged = tagged_smiles_from_tokens(
        product_tagged_tokens, product_untagged_tokens
    )

    n_new_tags = product_tagged_converted.count("!")

    if n_new_tags != n_tags:
        raise AssertionError(
            f"The number of tags is not the same after converting to '!' tagging. "
            f"product_tagged_atom_map: {product_atom_map_tagged}"
            f"product_tagged_converted: {product_tagged_converted}."
        )

    if product_tagged_converted.replace("!", "") != product_untagged:
        raise AssertionError(
            f"product_tagged.replace('!', '') != product_untagged."
            f"product_tagged: {product_tagged_converted}, product_untagged: {product_untagged}"
        )

    return product_tagged_converted



""" This module contains a collection of routines to produce pretty images
"""

from __future__ import annotations

import io
from typing import TYPE_CHECKING

from PIL import Image, ImageDraw
from rdkit import Chem
from rdkit.Chem import Draw

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports
    from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

    PilColor = Union[str, Tuple[int, int, int]]
    FrameColors = Optional[Dict[bool, PilColor]]
    from PIL.Image import Image as PilImage


def molecule_to_image(mol: Chem.rdchem.Mol, frame_color: PilColor, size: int = 300) -> PilImage:
    """
    Create a pretty image of a molecule,
    with a colored frame around it

    :param mol: the molecule
    :param frame_color: the color of the frame
    :param size: the size of the image
    :return: the produced image
    """
    img = Draw.MolToImage(mol, size=(size, size))
    cropped_img = crop_image(img)
    return draw_rounded_rectangle(cropped_img, frame_color)


def molecules_to_images(
    mols: Sequence[Chem.rdchem.Mol],
    frame_colors: Sequence[PilColor],
    size: int = 300,
    draw_kwargs: Dict[str, Any] = None,
) -> List[PilImage]:
    """
    Create pretty images of molecules with a colored frame around each one of them.

    The molecules will be resized to be of similar sizes.

    :param smiles_list: the molecules
    :param frame_colors: the color of the frame for each molecule
    :param size: the sub-image size
    :param draw_kwargs: additional keyword-arguments sent to `MolsToGridImage`
    :return: the produced images
    """
    draw_kwargs = draw_kwargs or {}
    all_mols = Draw.MolsToGridImage(mols, molsPerRow=len(mols), subImgSize=(size, size), **draw_kwargs)
    if not hasattr(all_mols, "crop"):  # Is not a PIL image
        fileobj = io.BytesIO(all_mols.data)
        all_mols = Image.open(fileobj)

    images = []
    for idx, frame_color in enumerate(frame_colors):
        image_obj = all_mols.crop((size * idx, 0, size * (idx + 1), size))
        image_obj = crop_image(image_obj)
        images.append(draw_rounded_rectangle(image_obj, frame_color))
    return images


def crop_image(img: PilImage, margin: int = 20) -> PilImage:
    """
    Crop an image by removing white space around it

    :param img: the image to crop
    :param margin: padding, defaults to 20
    :return: the cropped image
    """
    # pylint: disable=invalid-name
    # First find the boundaries of the white area
    x0_lim = img.width
    y0_lim = img.height
    x1_lim = 0
    y1_lim = 0
    for x in range(0, img.width):
        for y in range(0, img.height):
            if img.getpixel((x, y)) != (255, 255, 255):
                if x < x0_lim:
                    x0_lim = x
                if x > x1_lim:
                    x1_lim = x
                if y < y0_lim:
                    y0_lim = y
                if y > y1_lim:
                    y1_lim = y
    x0_lim = max(x0_lim, 0)
    y0_lim = max(y0_lim, 0)
    x1_lim = min(x1_lim + 1, img.width)
    y1_lim = min(y1_lim + 1, img.height)
    # Then crop to this area
    cropped = img.crop((x0_lim, y0_lim, x1_lim, y1_lim))
    # Then create a new image with the desired padding
    out = Image.new(
        img.mode,
        (cropped.width + 2 * margin, cropped.height + 2 * margin),
        color="white",
    )
    out.paste(cropped, (margin + 1, margin + 1))
    return out


def draw_rounded_rectangle(img: PilImage, color: PilColor, arc_size: int = 20) -> PilImage:
    """
    Draw a rounded rectangle around an image

    :param img: the image to draw upon
    :param color: the color of the rectangle
    :param arc_size: the size of the corner, defaults to 20
    :return: the new image
    """
    # pylint: disable=invalid-name
    x0, y0, x1, y1 = img.getbbox()
    x1 -= 1
    y1 -= 1
    copy = img.copy()
    draw = ImageDraw.Draw(copy)
    arc_size_half = arc_size // 2
    draw.arc((x0, y0, arc_size, arc_size), start=180, end=270, fill=color)
    draw.arc((x1 - arc_size, y0, x1, arc_size), start=270, end=0, fill=color)
    draw.arc((x1 - arc_size, y1 - arc_size, x1, y1), start=0, end=90, fill=color)
    draw.arc((x0, y1 - arc_size, arc_size, y1), start=90, end=180, fill=color)
    draw.line((x0 + arc_size_half, y0, x1 - arc_size_half, y0), fill=color)
    draw.line((x1, arc_size_half, x1, y1 - arc_size_half), fill=color)
    draw.line((arc_size_half, y1, x1 - arc_size_half, y1), fill=color)
    draw.line((x0, arc_size_half, x0, y1 - arc_size_half), fill=color)
    return copy


class RouteImageFactory:
    """
    Factory class for drawing a route

    :param route: the dictionary representation of the route
    :param in_stock_colors: the colors around molecules, defaults to {True: "green", False: "orange"}
    :param show_all: if True, also show nodes that are marked as hidden
    :param margin: the margin between images
    :param mol_size: the size of the molecule
    :param mol_draw_kwargs: additional arguments sent to the drawing routine
    :param replace_mol_func: an optional function to replace molecule images
    """

    def __init__(
        self,
        route: Dict[str, Any],
        in_stock_colors: FrameColors = None,
        show_all: bool = True,
        margin: int = 100,
        mol_size: int = 300,
        mol_draw_kwargs: Dict[str, Any] = None,
        replace_mol_func: Callable[[Dict[str, Any]], None] = None,
    ) -> None:
        in_stock_colors = in_stock_colors or {
            True: "green",
            False: "orange",
        }
        self.show_all: bool = show_all
        self.margin: int = margin

        self._stock_lookup: Dict[str, Any] = {}
        self._mol_lookup: Dict[str, Any] = {}
        self._extract_molecules(route)
        if replace_mol_func is not None:
            replace_mol_func(self._mol_lookup)
        images = molecules_to_images(
            list(self._mol_lookup.values()),
            [in_stock_colors[val] for val in self._stock_lookup.values()],
            size=mol_size,
            draw_kwargs=mol_draw_kwargs or {},
        )
        self._image_lookup = dict(zip(self._mol_lookup.keys(), images))

        self._mol_tree = self._extract_mol_tree(route)
        self._add_effective_size(self._mol_tree)

        pos0 = (
            self._mol_tree["eff_width"] - self._mol_tree["image"].width + self.margin,
            int(self._mol_tree["eff_height"] * 0.5) - int(self._mol_tree["image"].height * 0.5),
        )
        self._add_pos(self._mol_tree, pos0)

        self.image = Image.new(
            self._mol_tree["image"].mode,
            (self._mol_tree["eff_width"] + self.margin, self._mol_tree["eff_height"]),
            color="white",
        )
        self._draw = ImageDraw.Draw(self.image)
        self._make_image(self._mol_tree)
        self.image = crop_image(self.image)

    def _add_effective_size(self, tree_dict: Dict[str, Any]) -> None:
        children = tree_dict.get("children", [])
        for child in children:
            self._add_effective_size(child)
        if children:
            tree_dict["eff_height"] = sum(child["eff_height"] for child in children) + self.margin * (len(children) - 1)
            tree_dict["eff_width"] = (
                max(child["eff_width"] for child in children) + tree_dict["image"].size[0] + self.margin
            )
        else:
            tree_dict["eff_height"] = tree_dict["image"].size[1]
            tree_dict["eff_width"] = tree_dict["image"].size[0] + self.margin

    def _add_pos(self, tree_dict: Dict[str, Any], pos: Tuple[int, int]) -> None:
        tree_dict["left"] = pos[0]
        tree_dict["top"] = pos[1]
        children = tree_dict.get("children")
        if not children:
            return

        mid_y = pos[1] + int(tree_dict["image"].height * 0.5)  # Mid-point of image along y
        children_height = sum(child["eff_height"] for child in children) + self.margin * (len(children) - 1)
        childen_leftmost = min(pos[0] - self.margin - child["image"].width for child in children)
        child_y = mid_y - int(children_height * 0.5)  # Top-most edge of children
        child_ys = []
        # Now compute first guess of y-pos for children
        for child in children:
            y_adjust = int((child["eff_height"] - child["image"].height) * 0.5)
            child_ys.append(child_y + y_adjust)
            child_y += self.margin + child["eff_height"]

        for idx, (child, child_y0) in enumerate(zip(children, child_ys)):
            child_x = childen_leftmost  # pos[0] - self.margin - child["image"].width
            child_y = child_y0
            # Overwrite first guess if child does not have any children
            if not child.get("children") and idx == 0 and len(children) > 1:
                child_y = child_ys[idx + 1] - self.margin - child["image"].height
            elif not child.get("children") and idx > 0:
                child_y = child_ys[idx - 1] + self.margin + children[idx - 1]["image"].height
            self._add_pos(child, (child_x, child_y))

    def _extract_mol_tree(self, tree_dict: Dict[str, Any]) -> Dict[str, Any]:
        dict_ = {
            "smiles": tree_dict["smiles"],
            "image": self._image_lookup[tree_dict["smiles"]],
        }
        if tree_dict.get("children"):
            dict_["children"] = [
                self._extract_mol_tree(grandchild)
                for grandchild in tree_dict.get("children")[0]["children"]  # type: ignore
                if not (grandchild.get("hide", False) and not self.show_all)
            ]
        return dict_

    def _extract_molecules(self, tree_dict: Dict[str, Any]) -> None:
        if tree_dict["type"] == "mol":
            self._stock_lookup[tree_dict["smiles"]] = tree_dict.get("in_stock", False)
            self._mol_lookup[tree_dict["smiles"]] = Chem.MolFromSmiles(tree_dict["smiles"])
        for child in tree_dict.get("children", []):
            self._extract_molecules(child)

    def _make_image(self, tree_dict: Dict[str, Any]) -> None:
        self.image.paste(tree_dict["image"], (tree_dict["left"], tree_dict["top"]))
        children = tree_dict.get("children")
        if not children:
            return

        children_right = max(child["left"] + child["image"].width for child in children)
        mid_x = children_right + int(0.5 * (tree_dict["left"] - children_right))
        mid_y = tree_dict["top"] + int(tree_dict["image"].height * 0.5)

        self._draw.line((tree_dict["left"], mid_y, mid_x, mid_y), fill="black")
        for child in children:
            self._make_image(child)
            child_mid_y = child["top"] + int(0.5 * child["image"].height)
            self._draw.line(
                (
                    mid_x,
                    mid_y,
                    mid_x,
                    child_mid_y,
                    child["left"] + child["image"].width,
                    child_mid_y,
                ),
                fill="black",
            )
        self._draw.ellipse((mid_x - 8, mid_y - 8, mid_x + 8, mid_y + 8), fill="black", outline="black")



"""
Contains routines for creating, reading, and writing n-gram collections

Can be run as a module to create a collection from a set of routes:

    python -m rxnutils.routes.retro_bleu.ngram_collection --filename routes.json --output ngrams.json --nitems 2 --metadata template_hash

"""

from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Set, Tuple

from tqdm import tqdm

from rxnutils.routes.base import SynthesisRoute


@dataclass
class NgramCollection:
    """
    Class to create, read, and write a collection of n-grams

    :param nitems: the length of each n-gram
    :param metadata_key: the key used to extract the n-grams from the reactions
    :param ngrams: the extracted n-grams
    """

    nitems: int
    metadata_key: str
    ngrams: Set[Tuple[str, ...]]

    @classmethod
    def from_file(cls, filename: str) -> "NgramCollection":
        """
        Read an n-gram collection from a JSON-file

        :param filename: the path to the file
        :return: the n-gram collection
        """
        with open(filename, "r") as fileobj:
            dict_ = json.load(fileobj)
        ngrams = {tuple(item.split("\t")) for item in dict_["ngrams"]}
        return NgramCollection(dict_["nitems"], dict_["metadata_key"], ngrams)

    @classmethod
    def from_tree_collection(cls, filename: str, nitems: int, metadata_key: str) -> "NgramCollection":
        """
        Make a n-gram collection by extracting them from a collection of
        synthesis routes.

        :param filename: the path to a file with a list of synthesis routes
        :param nitems: the length of the gram
        :param metadata_key: the metadata to extract
        :return: the n-gram collection
        """
        with open(filename, "r") as fileobj:
            tree_list = json.load(fileobj)

        ngrams = set()
        for tree_dict in tqdm(tree_list, leave=False):
            route = SynthesisRoute(tree_dict)
            ngrams = ngrams.union(route.reaction_ngrams(nitems, metadata_key))
        return NgramCollection(nitems, metadata_key, ngrams)

    def save_to_file(self, filename: str) -> None:
        """
        Save an n-gram collection to a JSON-file

        :param filename: the path to the file
        """
        dict_ = {
            "nitems": self.nitems,
            "metadata_key": self.metadata_key,
            "ngrams": ["\t".join(ngram) for ngram in self.ngrams],
        }
        with open(filename, "w") as fileobj:
            json.dump(dict_, fileobj)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser("Tool for making n-gram collections from a set of synthesis routes")
    parser.add_argument("--filename", nargs="+", help="the path to the synthesis routes")
    parser.add_argument("--output", help="the path to the n-gram collection")
    parser.add_argument("--nitems", type=int, help="the length of the gram")
    parser.add_argument("--metadata", help="the reaction metadata to extract for making the n-grams")
    args = parser.parse_args()

    collection = None
    for filename in args.filename:
        temp = NgramCollection.from_tree_collection(filename, args.nitems, args.metadata)
        if collection is None:
            collection = temp
        else:
            collection.ngrams = collection.ngrams.union(temp.ngrams)
    print(f"Collected unique {len(collection.ngrams)} {args.nitems}-grams")
    collection.save_to_file(args.output)



""" 
Module containing routines for Deepset route scoring

This is an implementation of the code used in the following publication,
original implementation by Guo Yujia

Yujia G, Kabeshov M, Le THD, Genheden S, Bergonzini G, Engkvist O, et al. 
A Deep Learning with Expert Augmentation Approach for Route Scoring in Organic Synthesis. ChemRxiv. 2024; 
doi:10.26434/chemrxiv-2024-tp7rh 
"""

from typing import Dict

import numpy as np
from rdkit import Chem

from rxnutils.chem.features.sc_score import SCScore
from rxnutils.routes.base import SynthesisRoute
from rxnutils.routes.deepset.featurizers import collect_reaction_features, default_reaction_featurizer, ecfp_fingerprint

try:
    import onnxruntime
except ImportError:
    SCORE_SUPPORTED = False
else:
    SCORE_SUPPORTED = True


class DeepsetModelClient:
    """
    Interface for an in-memory instance of a Deepset model for route scoring

    :params model_path: the path to an ONNX model file
    """

    def __init__(self, model_path: str):
        if not SCORE_SUPPORTED:
            raise ImportError("Cannot score routes with Deepset model. Install rxnutils with optional dependencies")

        session_options = onnxruntime.SessionOptions()
        session_options.intra_op_num_threads = 1
        self._deepnet = onnxruntime.InferenceSession(model_path, sess_options=session_options)

    def __call__(self, reaction_features: np.ndarray, route_features: np.ndarray) -> float:
        """
        Computes the route score using reaction and route features

        :param reaction_features: the packed reaction features
        :param route_features: the route features
        """
        input_ = {
            "reaction_features": reaction_features.astype(np.float32),
            "route_features": route_features.astype(np.float32),
        }
        return float(self._deepnet.run(None, input_)[0])


def deepset_route_score(
    route: SynthesisRoute,
    model_client: DeepsetModelClient,
    scscorer: SCScore,
    reaction_class_ranks: Dict[str, int],
) -> float:
    """
    Scores a synthesis route using a Deepset model

    Currently, it uses defaults for featurizers as
    described in the original publication.

    :params route: the route to score
    :params model_client: the interface to the model
    :params scscorer: the interface to the SCScore
    :params reaction_class_ranks: the lookup from reaction class to rank
    :returns: the computed score
    """
    # To avoid circular imports
    from rxnutils.routes.scoring import badowski_route_score

    target_mol = Chem.MolFromSmiles(route.reaction_tree["smiles"])
    target_fp = ecfp_fingerprint(target_mol)

    rank_score, reaction_features = collect_reaction_features(
        route.reaction_data(),
        target_fp,
        reaction_class_ranks,
        default_reaction_featurizer,
    )
    cost_score = badowski_route_score(route)
    volume_score = sum(route.leaf_counts().values())
    complexity_score = _complexity_score(route, scscorer)
    route_features = np.asarray([cost_score, volume_score, complexity_score, rank_score])
    return model_client(reaction_features, route_features)


def _complexity_score(route: SynthesisRoute, scscorer: SCScore) -> float:
    """Sums up the SCScore of all the intermediates in the route"""
    score = 0
    for smiles, count in route.intermediate_counts().items():
        mol = Chem.MolFromSmiles(smiles)
        if not mol:
            continue
        scscore = scscorer(mol)
        score += count * scscore
    return score



"""Module containing script to atom-map USPTO or ORD reactions
"""

import argparse
from typing import Optional, Sequence

import pandas as pd

from rxnutils.data.batch_utils import read_csv_batch

try:
    from rxnmapper import BatchedMapper  # pylint: disable=all
except ImportError:
    DEPENDENCY_OK = False
else:
    DEPENDENCY_OK = True
    from transformers import logging

    logging.set_verbosity_error()


def main(input_args: Optional[Sequence[str]] = None) -> None:
    """Function for command-line tool"""

    if not DEPENDENCY_OK:
        raise ImportError("You need to run this tool in an environment where rxnmapper is installed")

    parser = argparse.ArgumentParser("Script to atom-map USPTO or ORD reactions")
    parser.add_argument("--input", default="data.csv", help="the file with reactions")
    parser.add_argument("--output", default="data_mapped.csv", help="the output filename")
    parser.add_argument(
        "--batch",
        type=int,
        nargs=2,
        help="Line numbers to start and stop reading rows",
    )
    parser.add_argument(
        "--column",
        default="ReactionSmilesClean",
        help="The column with the reaction SMILES",
    )
    parser.add_argument(
        "--mapper_batch_size",
        type=int,
        default=5,
        help="how many SMILES to mapped at the same time",
    )
    args = parser.parse_args(input_args)

    data = read_csv_batch(args.input, sep="\t", index_col=False, batch=args.batch)

    rxn_mapper = BatchedMapper(batch_size=args.mapper_batch_size)
    mapped_data = []
    # Get RXNMapper results in batches with error handling (returning and empty dict for failed reaction)
    for start in range(0, len(data), args.mapper_batch_size):
        batch = data[args.column].iloc[start : start + args.mapper_batch_size].tolist()
        batch_mapped = list(rxn_mapper.map_reactions_with_info(batch))
        # Check RXNMapper results and if there is a failure (empty dict) then use the original Reaction SMILES
        for idx, (original_rsmi, mapped_rsmi) in enumerate(zip(batch, batch_mapped)):
            # In case of failure return the original Reaction SMILES
            if not mapped_rsmi:
                batch_mapped[idx] = {
                    "mapped_rxn": original_rsmi,
                    "confidence": 0.0,
                }
        mapped_data.extend(batch_mapped)
    mapped_data_df = pd.DataFrame(mapped_data)

    data = data.assign(**{column: mapped_data_df[column] for column in mapped_data_df.columns})
    data.to_csv(args.output, sep="\t", index=False)


if __name__ == "__main__":
    main()



"""Module containing actions on reactions that doesn't modify the reactions only compute properties of them"""

from __future__ import annotations

import json
from collections import defaultdict
from dataclasses import dataclass
from typing import ClassVar, List, Optional, Set, Tuple

import pandas as pd
from rdkit import Chem, RDLogger
from rdkit.Chem import AllChem, Descriptors
from rdkit.Chem.MolKey.InchiInfo import InchiInfo
from rdkit.Chem.rdMolDescriptors import CalcNumRings
from rdkit.Chem.rdmolops import FindPotentialStereo

from rxnutils.chem.cgr import CondensedGraphReaction
from rxnutils.chem.reaction import ChemicalReaction
from rxnutils.chem.utils import (
    atom_mapping_numbers,
    has_atom_mapping,
    reaction_centres,
    split_rsmi,
    split_smiles_from_reaction,
)
from rxnutils.pipeline.base import ReactionActionMixIn, action, global_apply

rd_logger = RDLogger.logger()
rd_logger.setLevel(RDLogger.CRITICAL)


@action
@dataclass
class CountComponents(ReactionActionMixIn):
    """Action for counting reaction components"""

    pretty_name: ClassVar[str] = "count_components"
    in_column: str
    nreactants_column: str = "NReactants"
    nmapped_reactants_column: str = "NMappedReactants"
    nreagents_column: str = "NReagents"
    nmapped_reagents_column: str = "NMappedReagents"
    nproducts_column: str = "NProducts"
    nmapped_products_column: str = "NMappedProducts"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_data = global_apply(data, self._apply_row, axis=1)
        return data.assign(**{column: new_data[column] for column in new_data.columns})

    def _apply_row(self, row: pd.Series) -> pd.Series:
        def _process_smiles(smiles_list: List[str]) -> Tuple[int, int]:
            if not smiles_list:
                return 0, 0
            smiles_list_mapped = [smi for smi in smiles_list if has_atom_mapping(smi, sanitize=False)]
            return len(smiles_list), len(smiles_list_mapped)

        reactants_list, reagents_list, products_list = self.split_lists(row)
        nreactants, nmapped_reactants = _process_smiles(reactants_list)
        nreagents, nmapped_reagents = _process_smiles(reagents_list)
        nproducts, nmapped_products = _process_smiles(products_list)
        return pd.Series(
            {
                self.nreactants_column: nreactants,
                self.nmapped_reactants_column: nmapped_reactants,
                self.nreagents_column: nreagents,
                self.nmapped_reagents_column: nmapped_reagents,
                self.nproducts_column: nproducts,
                self.nmapped_products_column: nmapped_products,
            }
        )

    def __str__(self) -> str:
        return f"{self.pretty_name} (counting reactants, reagents, products and mapped versions of these)"


@action
@dataclass
class CountElements:
    """Action for counting elements in reactants"""

    pretty_name: ClassVar[str] = "count_elements"
    in_column: str
    out_column: str = "ElementCount"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        hash_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: hash_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (calculate the occurence of elements in the reactants)"

    def _row_action(self, row: pd.Series) -> str:
        def count_elements(smiles, counts):
            mol = Chem.MolFromSmiles(smiles)
            if mol is None:
                return
            for atom in mol.GetAtoms():
                counts[atom.GetAtomicNum()] += 1

        counts = defaultdict(int)
        reactants, _, _ = split_rsmi(row[self.in_column])
        for smiles in split_smiles_from_reaction(reactants):
            count_elements(smiles, counts)

        return json.dumps(counts)


@action
@dataclass
class HasStereoInfo:
    """Action for checking stereo info"""

    pretty_name: ClassVar[str] = "has_stereo_info"
    in_column: str
    out_column: str = "HasStereo"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (returns boolean if SMILES has stereo info (contains '@'))"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        return "@" in smiles


@action
@dataclass
class HasUnmappedRadicalAtom:
    """Action for flagging if reaction has any unmapped radical atoms"""

    pretty_name: ClassVar[str] = "hasunmappedradicalatom"
    in_column: str
    out_column: str = "HasUnmappedRadicalAtom"

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(df, self._row_action, axis=1)
        return df.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (detect if there is an unmapped radical in the reaction SMILES)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, product = split_rsmi(row["rsmi_processed"])

        react_mol = Chem.MolFromSmiles(reactants)
        if Descriptors.NumRadicalElectrons(react_mol) == 0:
            return False

        prod_mol = Chem.MolFromSmiles(product)
        unmapped_product_atom_element = None
        for atom in prod_mol.GetAtoms():
            if atom.GetAtomMapNum() == 0:
                unmapped_product_atom_element = atom.GetAtomicNum()
                break

        if unmapped_product_atom_element is None:
            return False

        has_radical = False
        for atom in react_mol.GetAtoms():
            if (
                atom.GetAtomMapNum() == 0
                and atom.GetNumRadicalElectrons() > 0
                and atom.GetAtomicNum() == unmapped_product_atom_element
            ):
                has_radical = True
                break
        return has_radical


@action
@dataclass
class HasUnsanitizableReactants:
    """Action for flagging if reaction has any unsanitizable reactants"""

    pretty_name: ClassVar[str] = "unsanitizablereactants"
    rsmi_column: str
    bad_columns: List[str]
    out_column: str = "HasUnsanitizableReactants"

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(df, self._row_action, axis=1)
        return df.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (detect if there is unsanitizable reactants)"

    def _row_action(self, row: pd.Series) -> str:
        unsanitize_atom_map_num = set()
        for column in self.bad_columns:
            if not isinstance(row[column], str):
                continue
            unsanitize_atom_map_num |= self._process_bad_column(row[column])

        if not unsanitize_atom_map_num:
            return False

        _, _, product = split_rsmi(row[self.rsmi_column])
        prod_atom_map_num = set(atom_mapping_numbers(product))
        return bool(prod_atom_map_num.intersection(unsanitize_atom_map_num))

    @staticmethod
    def _process_bad_column(smiles_list: str) -> Set[int]:
        unsanitize_atom_map_num = set()
        for smiles in smiles_list.split(","):
            mol = Chem.MolFromSmiles(smiles, sanitize=False)
            if not mol:
                continue
            unsanitize_atom_map_num |= {atom.GetAtomMapNum() for atom in mol.GetAtoms() if atom.GetAtomMapNum()}
        return unsanitize_atom_map_num


@action
@dataclass
class CgrCreated:
    """Action for determining if a CGR can be created from the reaction smiles"""

    in_column: str
    pretty_name: ClassVar[str] = "cgr_created"
    out_column: str = "CGRCreated"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        out_col = global_apply(data, self._create_cgr_column, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (flag if a CGR can be created for the reaction)"

    def _create_cgr_column(self, row: pd.Series) -> bool:
        rxn = ChemicalReaction(row[self.in_column], clean_smiles=False)
        try:
            _ = CondensedGraphReaction(rxn)
        except Exception:
            return False
        return True


@action
@dataclass
class CgrNumberOfDynamicBonds:
    """Action for calculating the number of dynamic bonds"""

    in_column: str
    pretty_name: ClassVar[str] = "cgr_dynamic_bonds"
    out_column: str = "NDynamicBonds"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        out_col = global_apply(data, self._process_row, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (number of dynamic bonds in the CGR)"

    def _process_row(self, row: pd.Series) -> Optional[int]:
        rxn = ChemicalReaction(row[self.in_column], clean_smiles=False)
        try:
            cgr = CondensedGraphReaction(rxn)
        except Exception:
            return None
        return cgr.bonds_changed


@action
@dataclass
class ProductAtomMappingStats:
    """Action for collecting statistics of product atom mapping"""

    pretty_name: ClassVar[str] = "product_atommapping_stats"
    in_column: str
    unmapped_column: str = "UnmappedProdAtoms"
    widow_column: str = "WidowAtoms"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_data = global_apply(data, self._row_action, axis=1)
        return data.assign(**{column: new_data[column] for column in new_data.columns})

    def __str__(self) -> str:
        return f"{self.pretty_name} (count number of number of unmapped and widow product atoms)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, products = split_rsmi(row[self.in_column])
        prod_mol = Chem.MolFromSmiles(products)
        nprod_atoms = prod_mol.GetNumAtoms() if prod_mol else 0

        prod_atommappings = set(atom_mapping_numbers(products))
        react_atommappings = []
        for smi in split_smiles_from_reaction(reactants):
            react_atommappings.extend(atom_mapping_numbers(smi))
        react_atommappings = set(react_atommappings)

        return pd.Series(
            {
                self.unmapped_column: nprod_atoms - len(prod_atommappings),
                self.widow_column: len(prod_atommappings - react_atommappings),
            }
        )


@action
@dataclass
class ProductSize:
    """Action for counting product size"""

    pretty_name: ClassVar[str] = "productsize"
    in_column: str
    out_column: str = "ProductSize"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (number of heavy atoms in product)"

    def _row_action(self, row: pd.Series) -> str:
        _, _, products = split_rsmi(row[self.in_column])
        products_mol = Chem.MolFromSmiles(products)

        if products_mol:
            product_atom_count = products_mol.GetNumHeavyAtoms()
        else:
            product_atom_count = 0

        return product_atom_count


@action
@dataclass
class PseudoReactionHash:
    """Action for creating a reaction hash based on InChI keys"""

    pretty_name: ClassVar[str] = "pseudo_reaction_hash"
    in_column: str
    out_column: str = "PseudoHash"
    no_reagents: bool = False

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        hash_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: hash_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (calculate hash based on InChI key of components)"

    def _row_action(self, row: pd.Series) -> str:
        def components_to_inchi(smiles):
            mols = [Chem.MolFromSmiles(smi) for smi in split_smiles_from_reaction(smiles)]
            return ".".join(sorted([Chem.MolToInchiKey(mol) for mol in mols if mol]))

        if self.no_reagents:
            reactants, _, products = split_rsmi(row[self.in_column])
            return ">>".join([components_to_inchi(smi) for smi in [reactants, products]])

        return ">".join(components_to_inchi(smi) for smi in split_rsmi(row[self.in_column]))


@action
@dataclass
class PseudoSmilesHash:
    """Action for creating a reaction hash based on SMILES"""

    pretty_name: ClassVar[str] = "pseudo_smiles_hash"
    in_column: str
    out_column: str = "PseudoSmilesHash"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        hash_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: hash_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (calculate hash based on InChI key of components of SMILES)"

    def _row_action(self, row: pd.Series) -> str:
        def components_to_inchi(smiles):
            mols = [Chem.MolFromSmiles(smi) for smi in split_smiles_from_reaction(smiles)]
            return ".".join(sorted([Chem.MolToInchiKey(mol) for mol in mols if mol]))

        return components_to_inchi(row[self.in_column])


@action
@dataclass
class ReactantProductAtomBalance:
    """Action for computing atom balance"""

    pretty_name: ClassVar[str] = "atombalance"
    in_column: str
    out_column: str = "RxnAtomBalance"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (product atom count minus reactants atom count)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, products = split_rsmi(row[self.in_column])

        reactants_mol = Chem.MolFromSmiles(reactants)
        if reactants_mol:
            reactant_atom_count = reactants_mol.GetNumHeavyAtoms()
        else:
            reactant_atom_count = 0

        products_mol = Chem.MolFromSmiles(products)
        if products_mol:
            product_atom_count = products_mol.GetNumHeavyAtoms()
        else:
            product_atom_count = 0

        return product_atom_count - reactant_atom_count


@action
@dataclass
class ReactantSize:
    """Action for counting reactant size"""

    pretty_name: ClassVar[str] = "reactantsize"
    in_column: str
    out_column: str = "ReactantSize"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (number of heavy atoms in reactant)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, _ = split_rsmi(row[self.in_column])
        reactants_mol = Chem.MolFromSmiles(reactants)

        if reactants_mol:
            reactant_atom_count = reactants_mol.GetNumHeavyAtoms()
        else:
            reactant_atom_count = 0

        return reactant_atom_count


@action
@dataclass
class MaxRingNumber:
    """
    Action for calculating the maximum number of rings in either the product or reactant
    For a reaction without reactants or products, it will return 0 to enable easy arithmetic comparison
    """

    pretty_name: ClassVar[str] = "maxrings"
    in_column: str
    out_column: str = "MaxRings"

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(df, self._row_action, axis=1)
        return df.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (maximum number of rings)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, products = split_rsmi(row[self.in_column])

        reactants_mols = [Chem.MolFromSmiles(smi) for smi in reactants.split(".")]
        nrings_reactants = [CalcNumRings(mol) for mol in reactants_mols if mol]
        max_rings_reactants = 0 if not nrings_reactants else max(nrings_reactants)

        prod_mol = Chem.MolFromSmiles(products)
        if not prod_mol:
            return max_rings_reactants

        nrings_product = CalcNumRings(prod_mol)
        return max([nrings_product, max_rings_reactants])


@action
@dataclass
class RingNumberChange:
    """
    Action for calculating if reaction has change in number of rings

    A positive number from this action implies that a ring was formed during the reaction
    """

    pretty_name: ClassVar[str] = "ringnumberchange"
    in_column: str
    out_column: str = "NRingChange"

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(df, self._row_action, axis=1)
        return df.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (ring change based on number of rings)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, products = split_rsmi(row[self.in_column])
        prod_mol = Chem.MolFromSmiles(products)

        if not prod_mol:
            return False

        reactants_mols = [Chem.MolFromSmiles(smi) for smi in reactants.split(".")]
        nrings_reactants = sum(CalcNumRings(mol) for mol in reactants_mols if mol)
        nrings_product = CalcNumRings(prod_mol)
        return nrings_product - nrings_reactants


@action
@dataclass
class RingBondMade:
    """Action for flagging if reaction has made a ring bond in the product"""

    pretty_name: ClassVar[str] = "ringbondmade"
    in_column: str
    out_column: str = "RingBondMade"

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(df, self._row_action, axis=1)
        return df.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (ring change based on ring bond made)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, products = split_rsmi(row[self.in_column])

        prod_mol = Chem.MolFromSmiles(products)
        if not prod_mol:
            return False
        ring_bonds = [
            (bond.GetBeginAtom().GetAtomMapNum(), bond.GetEndAtom().GetAtomMapNum())
            for bond in prod_mol.GetBonds()
            if bond.IsInRing()
        ]

        for smiles in split_smiles_from_reaction(reactants):
            reactant_mol = Chem.MolFromSmiles(smiles)
            if not reactant_mol:
                continue
            r_mappings = self._mapping_to_index(reactant_mol)
            for atom_map1, atom_map2 in ring_bonds:
                if atom_map1 not in r_mappings and atom_map2 not in r_mappings:
                    continue

                # If not both of the atoms are in the same molecule the bond is new
                if atom_map1 not in r_mappings and atom_map2 in r_mappings:
                    return True
                if atom_map2 not in r_mappings and atom_map1 in r_mappings:
                    return True

                atom_idx1 = r_mappings[atom_map1]
                atom_idx2 = r_mappings[atom_map2]
                bond = reactant_mol.GetBondBetweenAtoms(atom_idx1, atom_idx2)
                if bond is None or not bond.IsInRing():
                    return True
        return False

    @staticmethod
    def _mapping_to_index(mol):
        return {atom.GetAtomMapNum(): atom.GetIdx() for atom in mol.GetAtoms() if atom.GetAtomMapNum()}


@action
@dataclass
class RingMadeSize:
    """Action for computing the size of a newly formed ring"""

    pretty_name: ClassVar[str] = "ringmadesize"
    in_column: str
    out_column: str = "RingMadeSize"

    def __call__(self, df: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(df, self._row_action, axis=1)
        return df.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (largest ring made)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, _, products = split_rsmi(row[self.in_column])

        prod_mol = Chem.MolFromSmiles(products)
        if not prod_mol:
            return None

        reactant_rings = []
        for smiles in split_smiles_from_reaction(reactants):
            reactant_mol = Chem.MolFromSmiles(smiles)
            if not reactant_mol:
                continue

            reactant_rings.extend(self._find_rings(reactant_mol))

        largest_made_ring = 0
        for ring in self._find_rings(prod_mol):
            if ring not in reactant_rings:
                largest_made_ring = max(largest_made_ring, len(ring))
        return largest_made_ring

    @staticmethod
    def _find_rings(mol):
        rings = []
        for ring in mol.GetRingInfo().AtomRings():
            ring_atoms = [mol.GetAtomWithIdx(idx).GetAtomMapNum() for idx in ring]
            # Only consider at least partially atom-mapped rings
            if set(ring_atoms) == {0}:
                continue
            rings.append(tuple(sorted(ring_atoms)))
        return rings


@action
@dataclass
class SmilesLength:
    """Action for counting SMILES length"""

    pretty_name: ClassVar[str] = "smiles_length"
    in_column: str
    out_column: str = "SmilesLength"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (calculate length of smiles)"

    def _row_action(self, row: pd.Series) -> str:
        return len(row[self.in_column])


@action
@dataclass
class SmilesSanitizable:
    """Action for checking if SMILES are sanitizable"""

    pretty_name: ClassVar[str] = "smiles_sanitizable"
    in_column: str
    out_column: str = "SmilesSanitizable"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check if SMILES are sanitizable by RDKit)"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        if not smiles:
            return False

        mol = Chem.MolFromSmiles(smiles)
        try:  # Extra check that the molecule can generate SMILES that are again parsable
            mol = Chem.MolFromSmiles(Chem.MolToSmiles(mol))
        except Exception:  # pylint: disable=broad-except # noqa
            return False
        else:
            return bool(mol)


@action
@dataclass
class StereoInvention:
    """
    Flags reactions where non-stereo compounds (No "@"s in SMILES)
    turn into stereo compounds (containing "@")
    """

    pretty_name: ClassVar[str] = "stereo_invention"
    in_column: str
    out_column: str = "StereoInvention"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (Reactants with no stereo, turn into stereo compounds)"

    def _has_stereo(self, smiles: str) -> bool:
        return "@" in smiles

    def _row_action(self, row: pd.Series) -> bool:
        reactants, _, products = split_rsmi(row[self.in_column])

        return self._has_stereo(products) and not self._has_stereo(reactants)


@action
@dataclass
class StereoCentreChanges:
    """
    Action for checking if stereogenic centre in reaction center is changing
    during the reaction

    Will create two columns:
        1. A boolean column indicating True or False if it has stereochanges
        2. A description of the stereo information before and after the reaction
    """

    pretty_name: ClassVar[str] = "stereo_centre_changes"
    in_column: str
    out_column: str = "HasStereoChanges"
    stereo_changes_column: str = "StereoChanges"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_df = global_apply(data, self._row_action, axis=1)
        return data.assign(**{col: new_df[col] for col in new_df.columns})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check if stereochemistry changes in reaction)"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        rdkit_rxn = AllChem.ReactionFromSmarts(smiles, useSmiles=True)
        rdkit_rxn.Initialize()
        rxncenters = reaction_centres(rdkit_rxn)

        reactants = rdkit_rxn.GetReactants()
        reactants_chiral_flags = {}
        for centers, reactant in zip(rxncenters, reactants):
            for index in centers:
                atom = reactant.GetAtomWithIdx(index)
                reactants_chiral_flags[atom.GetAtomMapNum()] = str(atom.GetChiralTag())

        product_chiral_flags = {}
        for atom in rdkit_rxn.GetProducts()[0].GetAtoms():
            if atom.GetAtomMapNum() in reactants_chiral_flags:
                product_chiral_flags[atom.GetAtomMapNum()] = str(atom.GetChiralTag())

        if reactants_chiral_flags == product_chiral_flags:
            return pd.Series({self.out_column: False, self.stereo_changes_column: None})

        output = {}
        for atom_map, reactant_chirality in reactants_chiral_flags.items():
            product_chirality = product_chiral_flags.get(atom_map, None)
            if product_chirality != reactant_chirality:
                output[atom_map] = (reactant_chirality, product_chirality)
        return pd.Series({self.out_column: True, self.stereo_changes_column: json.dumps(output)})


@action
@dataclass
class StereoHasChiralReagent:
    """
    Action for checking if reagent has stereo centres
    """

    pretty_name: ClassVar[str] = "stereo_chiral_reagent"
    in_column: str
    out_column: str = "HasChiralReagent"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        if len(data) == 0:
            return data
        out_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check if reagent is chiral)"

    def _row_action(self, row: pd.Series) -> bool:
        smiles = row[self.in_column]
        return "@" in split_rsmi(smiles)[1]


@action
@dataclass
class StereoCenterIsCreated:
    """
    Action for checking if stereo centre is created during reaction
    """

    pretty_name: ClassVar[str] = "stereo_centre_created"
    in_column: str = "StereoChanges"
    out_column: str = "StereoCentreCreated"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        if len(data) == 0:
            return data
        out_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check if stereo centre is created)"

    def _row_action(self, row: pd.Series) -> str:
        try:
            chirality_changes = json.loads(row[self.in_column])
        except TypeError:
            return False
        for reactant_chirality, _ in chirality_changes.values():
            if reactant_chirality == "CHI_UNSPECIFIED":
                return True
        return False


@action
@dataclass
class StereoCenterIsRemoved:
    """
    Action for checking if stereo centre is removed during reaction
    """

    pretty_name: ClassVar[str] = "stereo_centre_removed"
    in_column: str = "StereoChanges"
    out_column: str = "StereoCentreRemoved"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        if len(data) == 0:
            return data
        out_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check if stereo centre is removed)"

    def _row_action(self, row: pd.Series) -> str:
        try:
            chirality_changes = json.loads(row[self.in_column])
        except TypeError:
            return False
        for _, product_chirality in chirality_changes.values():
            if product_chirality == "CHI_UNSPECIFIED":
                return True
        return False


@action
@dataclass
class StereoCenterInReactantPotential:
    """
    Action for checking if there is a potential stereo centre in the reaction

    Do not consider changes to bond stereochemistry
    """

    pretty_name: ClassVar[str] = "potential_stereo_center"
    in_column: str
    out_column: str = "PotentialStereoCentre"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        if len(data) == 0:
            return data
        out_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check for existence of potential stereo centre)"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        rdkit_rxn = AllChem.ReactionFromSmarts(smiles, useSmiles=True)
        rdkit_rxn.Initialize()
        rxncenters = reaction_centres(rdkit_rxn)
        reactants = rdkit_rxn.GetReactants()

        atom_maps = set()
        for centers, reactant0 in zip(rxncenters, reactants):
            reactant = self._get_clean_mol_copy(reactant0)
            stereo_info = FindPotentialStereo(reactant)
            if any(
                not str(info.type).startswith("Bond")
                and info.centeredOn in centers
                and str(reactant.GetAtomWithIdx(info.centeredOn).GetChiralTag()) == "CHI_UNSPECIFIED"
                for info in stereo_info
            ):
                return True
            for index in centers:
                atom = reactant0.GetAtomWithIdx(index)
                atom_maps.add(atom.GetAtomMapNum())

        product0 = rdkit_rxn.GetProducts()[0]
        product = self._get_clean_mol_copy(product0)
        stereo_info = FindPotentialStereo(product)
        if any(
            not str(info.type).startswith("Bond")
            and product0.GetAtomWithIdx(info.centeredOn).GetAtomMapNum() in atom_maps
            and str(product.GetAtomWithIdx(info.centeredOn).GetChiralTag()) == "CHI_UNSPECIFIED"
            for info in stereo_info
        ):
            return True

        return False

    @staticmethod
    def _get_clean_mol_copy(mol0):
        """
        Return a copy of the molecule
        that is sanitized and without atom-map numbers
        """
        mol = AllChem.rdchem.Mol(mol0)
        for atom in mol.GetAtoms():
            atom.SetAtomMapNum(0)
        AllChem.SanitizeMol(mol)
        return mol


@action
@dataclass
class StereoCenterOutsideReaction:
    """
    Action for checking if there is a stereo centre outside the reaction centre
    """

    pretty_name: ClassVar[str] = "stereo_centre_outside"
    in_column: str
    out_column: str = "StereoOutside"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        if len(data) == 0:
            return data
        out_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check for existence of chirality outside reaction centre)"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        rdkit_rxn = AllChem.ReactionFromSmarts(smiles, useSmiles=True)
        rdkit_rxn.Initialize()
        rxncenters = reaction_centres(rdkit_rxn)

        reactants = rdkit_rxn.GetReactants()
        for centers, reactant in zip(rxncenters, reactants):
            for index, atom in enumerate(reactant.GetAtoms()):
                if index in centers:
                    continue
                if str(atom.GetChiralTag()) != "CHI_UNSPECIFIED":
                    return True
        return False


@action
@dataclass
class StereoMesoProduct:
    """
    Action for checking if the product is a meso compound
    """

    pretty_name: ClassVar[str] = "meso_product"
    in_column: str
    out_column: str = "MesoProduct"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        if len(data) == 0:
            return data
        out_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: out_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (check for existence of meso compound)"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        rdkit_rxn = AllChem.ReactionFromSmarts(smiles, useSmiles=True)
        rdkit_rxn.Initialize()
        product_mol = rdkit_rxn.GetProducts()[0]
        AllChem.SanitizeMol(product_mol)
        info = InchiInfo(AllChem.MolToInchi(product_mol))
        return info.get_sp3_stereo()["main"]["non-isotopic"][2]



"""
Module containing pipeline for downloading, transforming and cleaning USPTO data
This needs to be run in an environment with rxnutils installed
"""

from pathlib import Path

from metaflow import step

from rxnutils.data.base_pipeline import DataPreparationBaseFlow
from rxnutils.data.uspto.combine import main as combine_uspto
from rxnutils.data.uspto.download import main as download_uspto


class UsptoDataPreparationFlow(DataPreparationBaseFlow):
    """Pipeline for download UPSTO source file, combining them and do some clean-up"""

    data_prefix = "uspto"

    @step
    def start(self):
        """Download USPTO data from Figshare"""
        download_uspto(
            [
                "--folder",
                self.folder,
            ]
        )
        self.next(self.combine_files)

    @step
    def combine_files(self):
        """Combine USPTO data files and add IDs"""
        combine_uspto(["--folder", self.folder])
        self.next(self.setup_cleaning)

    @step
    def setup_cleaning(self):
        """Setup cleaning"""
        self._setup_cleaning()
        self.next(self.do_cleaning, foreach="partitions")

    @step
    def do_cleaning(self):
        """Perform cleaning of data"""
        self._do_cleaning()
        self.next(self.join_cleaning)

    @step
    def join_cleaning(self, _):
        """Combined cleaned batches of data"""
        self._join_cleaning()
        self.next(self.end)

    @step
    def end(self):
        """Final step, just print information"""
        print(f"Processed file is locate here: {Path(self.folder) / 'uspto_data_cleaned.csv'}")


if __name__ == "__main__":
    UsptoDataPreparationFlow()



"""Module containing script to calculate RXNFP for a set of reactions
"""

import argparse
from typing import Optional, Sequence

import numpy as np
import pandas as pd

from rxnutils.data.batch_utils import read_csv_batch

try:
    from rxnfp.transformer_fingerprints import (  # pylint: disable=all
        RXNBERTFingerprintGenerator,
        get_default_model_and_tokenizer,
    )
except ImportError:
    DEPENDENCY_OK = False
else:
    DEPENDENCY_OK = True


def main(input_args: Optional[Sequence[str]] = None) -> None:
    """Function for command-line tool"""

    if not DEPENDENCY_OK:
        raise ImportError("You need to run this tool in an environment where rxnfp is installed")

    parser = argparse.ArgumentParser("Script to calculate RXNFP")
    parser.add_argument("--input", default="data.csv", help="the file with reactions")
    parser.add_argument("--output", default="data_rxnfp.npz", help="the output filename")
    parser.add_argument(
        "--column",
        default="reaction_smiles",
        help="The column with the reaction SMILES",
    )
    parser.add_argument(
        "--batch",
        type=int,
        nargs=2,
        help="Line numbers to start and stop reading rows",
    )
    parser.add_argument(
        "--fp_batch_size",
        type=int,
        default=10,
        help="how many SMILES to calculate at the same time",
    )
    args = parser.parse_args(input_args)

    data = read_csv_batch(args.input, sep="\t", index_col=False, batch=args.batch)

    model, tokenizer = get_default_model_and_tokenizer(force_no_cuda=True)
    rxnfp_generator = RXNBERTFingerprintGenerator(model, tokenizer, force_no_cuda=True)

    fingerprints = []
    for start in range(0, len(data), args.fp_batch_size):
        batch = data[args.column].iloc[start : start + args.fp_batch_size].tolist()
        # Adding zeros for all failed SMILES
        fail_data = [[0.0] * 256 for _ in range(args.fp_batch_size)]
        try:
            fps_batch = rxnfp_generator.convert_batch(batch)
        except ValueError as err:
            print(err)
            print("\n\n".join(batch))
            fingerprints.extend(fail_data)
        except RuntimeError as err:
            print(err)
            print("\n\n".join(batch))
            fingerprints.extend(fail_data)
        else:
            fingerprints.extend(fps_batch)

    fingerprints = np.asarray(fingerprints)
    np.savez(args.output, rxnfps=fingerprints)


if __name__ == "__main__":
    main()



"""Module containing actions on reactions that modify the reaction in some way"""

from __future__ import annotations

import os
import re
import subprocess
import sys
import tempfile
from dataclasses import dataclass, field
from pathlib import Path
from typing import ClassVar, List, Tuple

import pandas as pd
from rdkit import Chem, RDLogger
from rdkit.Chem import RDConfig

from rxnutils.chem.disconnection_sites.atom_map_tagging import atom_map_tag_products
from rxnutils.chem.disconnection_sites.tag_converting import convert_atom_map_tag
from rxnutils.chem.utils import (
    atom_mapping_numbers,
    desalt_molecules,
    join_smiles_from_reaction,
    neutralize_molecules,
    recreate_rsmi,
    remove_atom_mapping,
    split_rsmi,
    split_smiles_from_reaction,
)
from rxnutils.pipeline.base import ReactionActionMixIn, action, global_apply

CONTRIB_INSTALLED = os.path.exists(RDConfig.RDContribDir)
if CONTRIB_INSTALLED:
    # RDKit contrib is not default part of PYTHONPATH
    sys.path.append(RDConfig.RDContribDir)
    sys.path.append(f"{RDConfig.RDContribDir}/RxnRoleAssignment")  # to make Nadines code import "utils"
    # fmt: off
    from RxnRoleAssignment.identifyReactants import reassignReactionRoles  # pylint: disable=all # noqa

    # fmt: on

rd_logger = RDLogger.logger()
rd_logger.setLevel(RDLogger.CRITICAL)


@action
@dataclass
class DesaltMolecules(ReactionActionMixIn):
    """Action for desalting molecules"""

    pretty_name: ClassVar[str] = "desalt_molecules"
    in_column: str
    out_column: str = "RxnDesalted"
    keep_something: bool = False

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._apply_row, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def _apply_row(self, row: pd.Series) -> pd.Series:
        reactants_list, reagents_list, products_list = self.split_lists(row)
        reactants_list = desalt_molecules(reactants_list, keep_something=self.keep_something)
        if reagents_list:
            reagents_list = desalt_molecules(reagents_list, keep_something=self.keep_something)
        products_list = desalt_molecules(products_list, keep_something=self.keep_something)
        return self.join_lists(reactants_list, reagents_list, products_list)

    def __str__(self) -> str:
        return f"{self.pretty_name} (desalt molecules using RDKit SaltRemover)"


@action
@dataclass
class NameRxn:
    """Action for calling namrxn"""

    pretty_name: ClassVar[str] = "namerxn"
    in_column: str
    options: str = ""
    nm_rxn_column: str = "NextMoveRxnSmiles"
    nmc_column: str = "NMC"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        _, infile = tempfile.mkstemp(suffix=".smi")
        _, outfile = tempfile.mkstemp(suffix=".smi")
        data[self.in_column].to_csv(infile, index=False, header=False)
        subprocess.call(["namerxn"] + self.options.split() + [infile, outfile])
        if not os.path.exists(outfile) or os.path.getsize(outfile) == 0:
            raise FileNotFoundError("Could not produce namerxn output. Make sure 'namerxn' program is in path")
        # Set the data types to 'str' classification id must be a string
        namerxn_data = pd.read_csv(
            outfile,
            sep=" ",
            names=[self.nm_rxn_column, self.nmc_column],
            dtype=str,
        )
        return data.assign(**{col: namerxn_data[col] for col in namerxn_data.columns})

    def __str__(self) -> str:
        return f"{self.pretty_name} (running 'namerxn' to do atom mapping and classification)"


@action
@dataclass
class NeutralizeMolecules(ReactionActionMixIn):
    """Action for neutralizing molecules"""

    pretty_name: ClassVar[str] = "neutralize_molecules"
    in_column: str
    out_column: str = "RxnNeutralized"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._apply_row, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def _apply_row(self, row: pd.Series) -> pd.Series:
        reactants_list, reagents_list, products_list = self.split_lists(row)
        reactants_list = neutralize_molecules(reactants_list)
        if reagents_list:
            reagents_list = neutralize_molecules(reagents_list)
        products_list = neutralize_molecules(products_list)
        return self.join_lists(reactants_list, reagents_list, products_list)

    def __str__(self) -> str:
        return f"{self.pretty_name} (neutralize molecules using RDKit neutralizer)"


@action
@dataclass
class ReactantsToReagents(ReactionActionMixIn):
    """Action for converting reactants to reagents"""

    pretty_name: ClassVar[str] = "reactants2reagents"
    in_column: str
    out_column: str = "RxnSmilesWithTrueReagents"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (putting all non-reacting reactants as reagents)"

    def _row_action(self, row: pd.Series) -> str:
        reactants_list_old, reagents_list, products_list = self.split_lists(row)

        product_indices = {idx for smi in products_list for idx in atom_mapping_numbers(smi)}

        reactants_list = []
        for reactant in reactants_list_old:
            reactant_indices = atom_mapping_numbers(reactant)
            if not reactant_indices or not any(idx in product_indices for idx in reactant_indices):
                reagents_list.append(reactant)
            else:
                reactants_list.append(reactant)
        return self.join_lists(reactants_list, reagents_list, products_list)


@action
@dataclass
class ReagentsToReactants:
    """Action for converting reagents to reactants"""

    pretty_name: ClassVar[str] = "reagents2reactants"
    in_column: str
    out_column: str = "RxnSmilesAllReactants"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (putting all reagents to reactants)"

    def _row_action(self, row: pd.Series) -> str:
        reactants, reagents, products = split_rsmi(row[self.in_column])
        if reagents:
            reactants = ".".join([reactants, reagents])
        return reactants + ">>" + products


@action
@dataclass
class RemoveAtomMapping(ReactionActionMixIn):
    """Action for removing all atom mapping"""

    pretty_name: ClassVar[str] = "remove_atom_mapping"
    in_column: str
    out_column: str = "RxnSmilesNoAtomMap"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (remove atom mapping)"

    def _row_action(self, row: pd.Series) -> str:
        reactants_list, reagents_list, products_list = self.split_lists(row)
        reactants_list = [remove_atom_mapping(smi, sanitize=False) for smi in reactants_list]
        reagents_list = [remove_atom_mapping(smi, sanitize=False) for smi in reagents_list]
        products_list = [remove_atom_mapping(smi, sanitize=False) for smi in products_list]

        return self.join_lists(reactants_list, reagents_list, products_list)


@action
@dataclass
class RemoveStereoInfo:
    """Action for removing stero information"""

    pretty_name: ClassVar[str] = "remove_stereo_info"
    in_column: str
    out_column: str = "RxnSmilesNoStereo"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (remove stereo information from the SMILES)"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        smiles_nostereo = smiles.replace("@", "")
        return smiles_nostereo


@action
@dataclass
class InvertStereo:
    """Action for inverting stero information"""

    pretty_name: ClassVar[str] = "invert_stereo"
    in_column: str
    out_column: str = "RxnSmilesInvertedStereo"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (invert stereo in SMILES (@@ => @ and @ => @@))"

    def _row_action(self, row: pd.Series) -> str:
        smiles = row[self.in_column]
        smiles_temp = smiles.replace("@@", "£").replace("@", "$")
        return smiles_temp.replace("$", "@@").replace("£", "@")


@action
@dataclass
class IsotopeInfo:
    """Action creating and modifying isotope information"""

    pretty_name: ClassVar[str] = "isotope_info"
    in_column: str
    isotope_column: str = "Isotope"
    out_column: str = "RxnSmilesWithoutIsotopes"
    match_regex: str = r"\[(?P<mass>[0-9]+)(?P<symbol>[A-Za-z][a-z]*)(?P<spec>[^\]]+)*\]"
    sub_regex: str = r"[\g<symbol>\g<spec>]"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        isotop_info = data[self.in_column].str.extract(self.match_regex)
        isotop_col = isotop_info["mass"] + isotop_info["symbol"]
        rsmi_col = data[self.in_column].str.replace(
            self.match_regex, self.sub_regex, regex=True
        )
        return data.assign(**{self.isotope_column: isotop_col, self.out_column: rsmi_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (extract and remove isotope information from reactions)"


@action
@dataclass
class RemoveExtraAtomMapping(ReactionActionMixIn):
    """Action for removing extra atom mapping"""

    pretty_name: ClassVar[str] = "remove_extra_atom_mapping"
    in_column: str
    out_column: str = "RxnSmilesReassignedAtomMap"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (removing atom maps in reactants and reagents not in products)"

    def _row_action(self, row: pd.Series) -> str:
        def reassign_atom_mapping(smiles: str) -> str:
            mol = Chem.MolFromSmiles(smiles)
            if mol is None:
                return smiles
            for atom in mol.GetAtoms():
                if atom.GetAtomMapNum() not in product_indices:
                    atom.SetAtomMapNum(0)
            return Chem.MolToSmiles(mol)

        reactants_list, reagents_list, products_list = self.split_lists(row)

        product_indices = {idx for smi in products_list for idx in atom_mapping_numbers(smi)}

        reactants_list = [reassign_atom_mapping(smi) for smi in reactants_list]
        reagents_list = [reassign_atom_mapping(smi) for smi in reagents_list]
        return self.join_lists(reactants_list, reagents_list, products_list)


@action
@dataclass
class RemoveUnchangedProducts(ReactionActionMixIn):
    """Compares the products with the reagents and reactants and remove unchanged products.

    Protonation is considered a difference, As example, if there's a HCl in the reagents and
    a Cl- in the products, it will not be removed.
    """

    pretty_name: ClassVar[str] = "remove_unchanged_products"
    in_column: str
    out_column: str = "RxnNoUnchangedProd"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (Remove unchanged products)"

    def _row_action(self, row: pd.Series) -> str:
        reactants_list, reagents_list, products_list = self.split_lists(row)

        reactants_mols = [Chem.MolFromSmiles(m) for m in reactants_list]
        reagents_mols = [Chem.MolFromSmiles(m) for m in reagents_list]
        products_mols = [Chem.MolFromSmiles(m) for m in products_list]

        new_products = []
        for product, product_smi in zip(products_mols, products_list):
            matches = [self._is_equal(product, r) for r in reactants_mols + reagents_mols]
            if not any(matches):
                new_products.append(product_smi)

        return self.join_lists(reactants_list, reagents_list, new_products)

    @staticmethod
    def _is_equal(mol1, mol2) -> bool:
        try:
            inchi1 = Chem.MolToInchiKey(mol1)
            inchi2 = Chem.MolToInchiKey(mol2)
        except Exception:  # pylint: disable=broad-except # noqa
            return False
        else:
            return inchi1 == inchi2


@action
@dataclass
class RemoveUnsanitizable(ReactionActionMixIn):
    """Action for removing unsanitizable reactions"""

    pretty_name: ClassVar[str] = "remove_unsanitizable"
    in_column: str
    out_column: str = "RxnSanitizable"
    bad_column: str = "BadMolecules"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_data = global_apply(data, self._apply_row, axis=1)
        return data.assign(
            **{
                self.out_column: new_data.rxn_smi,
                self.bad_column: new_data.bad_smiles,
            }
        )

    def _apply_row(self, row: pd.Series) -> pd.Series:
        def process_smiles(smiles_list: List[str]) -> Tuple[List[str], List[str]]:
            good_list = []
            bad_list = []
            for smiles in smiles_list:
                mol = Chem.MolFromSmiles(smiles)
                if mol:
                    try:
                        mol = Chem.MolFromSmiles(Chem.MolToSmiles(mol))
                    except Exception:  # pylint: disable=broad-except # noqa
                        bad_list.append(smiles)
                    else:
                        if mol:
                            good_list.append(smiles)
                        else:
                            bad_list.append(smiles)
                else:
                    bad_list.append(smiles)
            return good_list, bad_list

        reactants_list, reagents_list, products_list = self.split_lists(row)
        good_reactants, bad_reactants = process_smiles(reactants_list)
        good_reagents, bad_reagents = process_smiles(reagents_list)
        good_products, bad_products = process_smiles(products_list)

        bad_smiles = ",".join(bad_reactants + bad_reagents + bad_products)
        rxn_smi = self.join_lists(good_reactants, good_reagents, good_products)
        return pd.Series({"rxn_smi": rxn_smi, "bad_smiles": bad_smiles})

    def __str__(self) -> str:
        return f"{self.pretty_name} (removing molecules that is not sanitizable by RDKit)"


@action
@dataclass
class RDKitRxnRoles:
    """Action for assigning roles based on RDKit algorithm"""

    pretty_name: ClassVar[str] = "rdkit_RxnRoleAssignment"
    in_column: str
    out_column: str = "RxnRoleAssigned"

    def __post_init__(self):
        if not CONTRIB_INSTALLED:
            raise ImportError("This action cannot be used because the RDKit Contrib folder is not installed")

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:

        smiles_col = global_apply(data, lambda row: reassignReactionRoles(row[self.in_column]), axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (reaction role assignment using RDKit contrib (Nadine and Greg's))"


# fmt: off
_MAP_SCRIPT = Path(__file__).parent.parent.parent / "data" / "mapping.py"
# fmt: on


@action
@dataclass
class RxnMapper:
    """Action for mapping reactions with the RXNMapper tool"""

    pretty_name: ClassVar[str] = "rxnmapper"
    in_column: str
    out_column: str = "RxnmapperRxnSmiles"
    rxnmapper_command: str = "conda run -p ${RXNMAPPER_ENV_PATH} python " + str(_MAP_SCRIPT)

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_column = self._do_action(data[self.in_column])
        return data.assign(**{self.out_column: new_column})

    def _do_action(self, column: pd.Series) -> pd.Series:
        input_handle, input_path = tempfile.mkstemp(suffix=".csv")
        output_handle, output_path = tempfile.mkstemp(suffix=".csv")
        # Recreate Reaction SMILES by removing intra-molecular complexes
        tmp_column = column.apply(recreate_rsmi)
        tmp_column.to_csv(input_path, sep="\t", index=False)

        cmd = self.rxnmapper_command
        if "CONDA_PATH" in os.environ:
            cmd = os.environ["CONDA_PATH"] + os.sep + cmd
        arglist = cmd.split() + [
            "--input",
            input_path,
            "--column",
            self.in_column,
            "--output",
            output_path,
        ]
        _ = subprocess.check_output(arglist).decode("utf-8")

        mapped_data = pd.read_csv(output_path, sep="\t")

        os.remove(input_path)
        os.remove(output_path)
        os.close(input_handle)
        os.close(output_handle)

        return mapped_data["mapped_rxn"]

    def __str__(self) -> str:
        return f"{self.pretty_name} (RFP by rxnfp)"


@action
@dataclass
class SplitReaction(ReactionActionMixIn):
    """Action for splitting reaction into components"""

    pretty_name: ClassVar[str] = "split_reaction"
    in_column: str
    out_columns: list[str]

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        new_columns = global_apply(data, self._apply_row, axis=1)
        if len(self.out_columns) == 2:
            return data.assign(
                **{
                    self.out_columns[0]: new_columns.reactants,
                    self.out_columns[1]: new_columns.products,
                }
            )
        if len(self.out_columns) == 3:
            return data.assign(
                **{
                    self.out_columns[0]: new_columns.reactants,
                    self.out_columns[1]: new_columns.reagents,
                    self.out_columns[2]: new_columns.products,
                }
            )
        raise ValueError(f"Don't know what to return with {len(self.out_columns)} output columns")

    def __str__(self) -> str:
        return f"{self.pretty_name} (rxn => reactants, [reagents], products)"

    def _apply_row(self, row: pd.Series) -> pd.Series:
        reactants, reagents, products = self.split_smiles(row)
        return pd.Series({"reactants": reactants, "reagents": reagents, "products": products})


@action
@dataclass
class AtomMapTagDisconnectionSite(ReactionActionMixIn):
    """Action for tagging disconnection site in products with atom-map '[<atom>:1]'."""

    pretty_name: ClassVar[str] = "atom_map_tag_disconnection_site"
    in_column: str = "RxnSmilesClean"
    out_column: str = "products_atom_map_tagged"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(data, self._row_action, axis=1)
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (tag disconnection sites in products with '[<atom>:1]')"

    def _row_action(self, row: pd.Series) -> str:
        return atom_map_tag_products(row[self.in_column])


@action
@dataclass
class ConvertAtomMapDisconnectionTag(ReactionActionMixIn):
    """Action for converting atom-map tagging to exclamation mark tagging.

    yaml example:

    convert_atom_map_disconnection_tag:
        in_column_tagged: products_atom_map_tagged
        in_column_untagged: products
        out_column_tagged: products_tagged
        out_column_reconstructed: products_reconstructed
    """

    pretty_name: ClassVar[str] = "convert_atom_map_disconnection_tag"
    in_column: str = "products_atom_map_tagged"
    out_column_tagged: str = "products_tagged"
    out_column_reconstructed: str = "products_reconstructed"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_tagged_col = global_apply(data, self._row_action, axis=1)
        smiles_reconstructed_col = smiles_tagged_col.str.replace("!", "")

        return data.assign(
            **{
                self.out_column_tagged: smiles_tagged_col,
                self.out_column_reconstructed: smiles_reconstructed_col,
            }
        )

    def __str__(self) -> str:
        return f"{self.pretty_name} (convert disconnection tagging '[<atom>:1]' to '<atom>!')"

    def _row_action(self, row: pd.Series) -> str:
        product_tagged = convert_atom_map_tag(row[self.in_column])
        return product_tagged


@action
@dataclass
class TrimRxnSmiles:
    """Action from trimming reaction SMILES"""

    pretty_name: ClassVar[str] = "trim_rxn_smiles"
    in_column: str
    out_column: str = "RxnSmiles"
    smiles_column_index: int = 0

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        smiles_col = global_apply(
            data,
            lambda row: row[self.in_column].split(" ")[self.smiles_column_index],
            axis=1,
        )
        return data.assign(**{self.out_column: smiles_col})

    def __str__(self) -> str:
        return f"{self.pretty_name} (removing junk from the reaction smiles)"



"""
Module containing pipeline for extracting, transforming and cleaning Open reaction database data
This needs to be run in an environment with rxnutils installed
"""

import glob
import os
from pathlib import Path

from metaflow import Parameter, step

from rxnutils.data.base_pipeline import DataPreparationBaseFlow
from rxnutils.data.ord.import_ord_dataset import main as import_data


class OrdDataPreparationFlow(DataPreparationBaseFlow):
    """Pipeline for extracting ORD data and do some clean-up"""

    ord_data = Parameter("ord-data")

    data_prefix = "ord"

    @step
    def start(self):
        """Import ORD data"""
        output_name = str(Path(self.folder) / "ord_data.csv")
        if not Path(output_name).exists():
            filenames = glob.glob(os.path.join(self.ord_data, "data", "*", "*"))
            import_data(
                [
                    "--output",
                    output_name,
                    "--filenames",
                ]
                + filenames
            )
        self.next(self.setup_cleaning)

    @step
    def setup_cleaning(self):
        """Setup cleaning"""
        self._setup_cleaning()
        self.next(self.do_cleaning, foreach="partitions")

    @step
    def do_cleaning(self):
        """Perform cleaning of data"""
        self._do_cleaning()
        self.next(self.join_cleaning)

    @step
    def join_cleaning(self, _):
        """Combined cleaned batches of data"""
        self._join_cleaning()
        self.next(self.end)

    @step
    def end(self):
        """Final step, just print information"""
        print(f"Processed file is locate here: {Path(self.folder) / 'ord_data_cleaned.csv'}")


if __name__ == "__main__":
    OrdDataPreparationFlow()



""" Contains routine to score routes according to Retro-BLEU paper
"""

import numpy as np

from rxnutils.routes.base import SynthesisRoute
from rxnutils.routes.retro_bleu.ngram_collection import NgramCollection


def ngram_overlap_score(
    route: SynthesisRoute,
    ref: NgramCollection,
) -> float:
    """
    Calculate the fractional n-gram overlap of the n-grams in the given
    route and the reference n-gram collection

    :param route: the route to score
    :param ref: the reference n-gram collection
    :return: the calculated score
    """
    route_ngrams = set(route.reaction_ngrams(ref.nitems, ref.metadata_key))
    if not route_ngrams:
        return np.nan
    return len(route_ngrams.intersection(ref.ngrams)) / len(route_ngrams)


def retro_bleu_score(route: SynthesisRoute, ref: NgramCollection, ideal_steps: int = 3) -> float:
    """
    Calculate the Retro-BLEU score according to the paper:

    Li, Junren, Lei Fang, och Jian-Guang Lou.
    ”Retro-BLEU: quantifying chemical plausibility of retrosynthesis routes through reaction template sequence analysis”.
    Digital Discovery 3, nr 3 (2024): 482–90. https://doi.org/10.1039/D3DD00219E.

    :param route: the route to score
    :param ref: the reference n-gram collection
    :param ideal_steps: a length-penalization hyperparameter (see Eq 2 in ref)
    :return: the calculated score
    """
    overlap = ngram_overlap_score(route, ref)

    nreactions = len(route.reaction_smiles())
    heuristic_score = ideal_steps / max(nreactions, ideal_steps)

    return np.exp(overlap) + np.exp(heuristic_score)



"""
Code for curating USPTO yields.

Inspiration from this code: https://github.com/DocMinus/Yield_curation_USPTO

This could potentially be an action, but since it only make sens to use it
with USPTO data, it resides here for now.
"""

from dataclasses import dataclass

import numpy as np
import pandas as pd


@dataclass
class UsptoYieldCuration:
    """
    Action for curating USPTO yield columns
    """

    text_yield_column: str = "TextMinedYield"
    calc_yield_column: str = "CalculatedYield"
    out_column: str = "CuratedYield"

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        calc_yield = data[self.calc_yield_column].str.rstrip("%")
        calc_yield = pd.to_numeric(calc_yield, errors="coerce")
        calc_yield[(calc_yield < 0) | (calc_yield > 100)] = np.nan

        text_yield = data[self.text_yield_column].str.lstrip("~")
        text_yield = text_yield.str.rstrip("%")
        text_yield = text_yield.str.replace(">=", "", regex=False)
        text_yield = text_yield.str.replace(">", "", regex=False)
        text_yield = text_yield.str.replace("<", "", regex=False)
        text_yield = text_yield.str.replace(r"\d{1,2}\sto\s", "", regex=True)
        text_yield = pd.to_numeric(text_yield, errors="coerce")
        text_yield[(text_yield < 0) | (text_yield > 100)] = np.nan

        curated_yield = text_yield.copy()

        sel = (~calc_yield.isna()) & (~text_yield.isna())
        curated_yield[sel] = np.maximum(calc_yield[sel], text_yield[sel])

        sel = (~calc_yield.isna()) & (text_yield.isna())
        curated_yield[sel] = calc_yield[sel]

        return data.assign(**{self.out_column: curated_yield})

    def __str__(self) -> str:
        return f"{self.pretty_name} (create one column with curated yield values)"



"""Module containing a script to download USPTO files Figshare
"""

import argparse
import os
from pathlib import Path
from typing import Optional, Sequence

import py7zr
import requests
import tqdm

FILES_TO_DOWNLOAD = [
    {
        "filename": "2001_Sep2016_USPTOapplications_smiles.rsmi.7z",
        "url": "https://ndownloader.figshare.com/files/8664370",
    },
    {
        "filename": "1976_Sep2016_USPTOgrants_smiles.rsmi.7z",
        "url": "https://ndownloader.figshare.com/files/8664379",
    },
]


def _download_file(url: str, filename: str) -> None:
    with requests.get(url, stream=True) as response:
        response.raise_for_status()
        total_size = int(response.headers.get("content-length", 0))
        pbar = tqdm.tqdm(total=total_size, desc=os.path.basename(filename), unit="B", unit_scale=True)
        with open(filename, "wb") as fileobj:
            for chunk in response.iter_content(chunk_size=1024):
                fileobj.write(chunk)
                pbar.update(len(chunk))
        pbar.close()


def main(args: Optional[Sequence[str]] = None) -> None:
    """Function for command-line tool"""
    parser = argparse.ArgumentParser("Script to download USPTO SMILES data from Figshare")
    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="if given will overwrite existing files",
    )
    parser.add_argument("--folder", default=".", help="folder for downloading")
    args = parser.parse_args(args)

    for file_spec in FILES_TO_DOWNLOAD:
        filename = Path(args.folder) / file_spec["filename"]
        if not args.overwrite and filename.exists():
            continue
        _download_file(file_spec["url"], filename)

        archive = py7zr.SevenZipFile(filename, mode="r")
        archive.extractall(path=args.folder)
        archive.close()


if __name__ == "__main__":
    main()



""" Wrapper class for the CGRTools library
"""

import io
import warnings
from typing import List

from CGRtools.containers.molecule import MoleculeContainer
from CGRtools.containers.reaction import ReactionContainer
from CGRtools.files.SDFrw import SDFRead
from rdkit import Chem

from rxnutils.chem.reaction import ChemicalReaction
from rxnutils.chem.utils import atom_mapping_numbers


class CondensedGraphReaction:
    """
    The Condensed Graph of Reaction (CGR) representation of a reaction

    :ivar reaction_container: the CGRTools container of the reaction
    :ivar cgr_container: the CGRTools container of the CGR
    :param reaction: the reaction composed of RDKit molecule to start from
    :raises ValueError: if it is not possible to create the CGR from the reaction
    """

    def __init__(self, reaction: ChemicalReaction) -> None:
        self.reaction = reaction
        self._cgr_reactants = []
        self._cgr_products = []
        self._make_cgr_containers()
        self.reaction_container = ReactionContainer(reactants=self._cgr_reactants, products=self._cgr_products)
        try:
            self.cgr_container = self.reaction_container.compose()
        except ValueError as err:
            if str(err) == "mapping of graphs is not disjoint":
                raise ValueError("Reaction contains inconsistent atom-mapping, perhaps duplicates")
            elif str(err).endswith("} not equal"):
                raise ValueError("Atom with the same atom-mapping in reactant and product is not equal")
            else:
                raise ValueError(f"Unknown problem with generating CGR: {err}")

    def __eq__(self, obj: object) -> bool:
        if not isinstance(obj, CondensedGraphReaction):
            return False
        return self.cgr_container == obj.cgr_container

    @property
    def bonds_broken(self) -> int:
        """Returns the number of broken bonds in the reaction"""
        return sum(bond.p_order is None for _, _, bond in self.cgr_container.bonds())

    @property
    def bonds_changed(self) -> int:
        """Returns the number of broken or formed bonds in the reaction"""
        return sum(bond.p_order is None or bond.order is None for _, _, bond in self.cgr_container.bonds())

    @property
    def bonds_formed(self) -> int:
        """Returns the number of formed bonds in the reaction"""
        return sum(bond.order is None for _, _, bond in self.cgr_container.bonds())

    @property
    def total_centers(self) -> int:
        """Returns the number of atom and bond centers in the reaction"""
        return len(self.cgr_container.center_atoms) + len(self.cgr_container.center_bonds)

    def distance_to(self, other: "CondensedGraphReaction") -> int:
        """
        Returns the chemical distance between two reactions, i.e. the absolute difference
        between the total number of centers.

        Used for some atom-mapping comparison statistics

        :param other: the reaction to compare to
        :returns: the computed distance
        """
        return abs(self.total_centers - other.total_centers)

    def _make_cgr_containers(self):
        nreactants = len(self.reaction.reactants)
        renumbered_mols = self._make_renumbered_mols()
        self._cgr_reactants = _convert_rdkit_molecules(renumbered_mols[:nreactants])
        self._cgr_products = _convert_rdkit_molecules(renumbered_mols[nreactants:])

        if len(self._cgr_reactants) != nreactants:
            warnings.warn("Warning not all reactants could be converted to CGRtools")
        if len(self._cgr_products) != len(self.reaction.products):
            warnings.warn("Warning not all products could be converted to CGRtools")

    def _make_renumbered_mols(self):
        # It is necessary that all atoms have atom-mapping
        # otherwise CGRTools will start adding bad atom-mappings
        # so this adds safe atom-mapping to un-mapped atoms
        renumbered_mols = []
        max_atom_map_numb = max(
            max(atom_mapping_numbers(smi) or [0]) for smi in self.reaction.reactants_list + self.reaction.products_list
        )
        for mol0 in self.reaction.reactants + self.reaction.products:
            if mol0 is None:
                raise ValueError("Cannot create CGR for this reaction, some molecules are None")
            mol = Chem.rdchem.Mol(mol0)
            for atom in mol.GetAtoms():
                if not atom.GetAtomMapNum():
                    max_atom_map_numb += 1
                    atom.SetAtomMapNum(max_atom_map_numb)
            renumbered_mols.append(mol)
        return renumbered_mols


def _convert_rdkit_molecules(
    mols: List[Chem.rdchem.Mol],
) -> List[MoleculeContainer]:
    # This connects the SDWriter of RDKit with the SDFRead of CGRTools
    with io.StringIO() as sd_buffer:
        with Chem.SDWriter(sd_buffer) as sd_writer_obj:
            for mol in mols:
                sd_writer_obj.write(mol)
        sd_buffer.seek(0)
        with SDFRead(sd_buffer, remap=False) as sd_reader_obj:
            cgr_molecules = sd_reader_obj.read()

    # This make the aromatization of the molecules, which were
    # lost when writing to SD format. This is essential to keep
    # the reactant and product structures as equal as possible
    for mol in cgr_molecules:
        mol.thiele()
    return cgr_molecules



"""Module containing a class to handle chemical reactions"""

import hashlib
from typing import Any, Dict, List, Optional, Tuple

import wrapt_timeout_decorator
from rdchiral import template_extractor as extractor
from rdkit import Chem
from rdkit.Chem import AllChem

from rxnutils.chem import utils
from rxnutils.chem.rinchi import rinchi_api
from rxnutils.chem.template import ReactionTemplate
from rxnutils.chem.utils import reassign_rsmi_atom_mapping, split_rsmi, split_smiles_from_reaction


class ReactionException(Exception):
    """Custom exception raised when failing operations on a chemical reaction"""


class ChemicalReaction:
    """
    Representation of chemical reaction

    :param smiles: the reaction SMILES
    :param id_: an optional database ID of the reaction
    :param clean_smiles: if True, will standardize the reaction SMILES
    """

    def __init__(self, smiles: str, id_: str = None, clean_smiles: bool = True) -> None:
        self.rsmi = smiles
        self.clean_rsmi = ""
        self.rid = id_

        self.agents_smiles: str = ""
        self.products_smiles: str = ""
        self.reactants_smiles: str = ""
        if clean_smiles:
            self.rsmi = reassign_rsmi_atom_mapping(self.rsmi, as_smiles=True)
            self._clean_reaction()
        else:
            self._split_reaction()

        self.reactants: List[Chem.rdchem.Mol] = extractor.mols_from_smiles_list(self.reactants_list)
        self.agents: List[Chem.rdchem.Mol] = extractor.mols_from_smiles_list(self.agents_list)
        self.products: List[Chem.rdchem.Mol] = extractor.mols_from_smiles_list(self.products_list)

        self._pseudo_rinchi: Optional[str] = None
        self._pseudo_rinchi_key: Optional[str] = None
        self._rinchi_data: Optional[rinchi_api.RInChIStructure] = None

        self._canonical_template: Optional[ReactionTemplate] = None
        self._retro_template: Optional[ReactionTemplate] = None

    @property
    def agents_list(self) -> List[str]:
        """Gives all the agents as strings"""
        return split_smiles_from_reaction(self.agents_smiles)

    @property
    def canonical_template(self) -> ReactionTemplate:
        """Gives the canonical (forward) template"""
        if self._canonical_template is None:
            raise ValueError(
                f"Call {self.__class__.__name__}.generate_reaction_template() "
                "before using {self.__class__.__name__}.canonical_template"
            )
        return self._canonical_template

    @property
    def products_list(self) -> List[str]:
        """Gives all products as strings"""
        return split_smiles_from_reaction(self.products_smiles)

    @property
    def pseudo_rinchi(self) -> str:
        """Gives pseudo RInChI"""
        if self._pseudo_rinchi is None:
            self._pseudo_rinchi = self._make_pseudo_rinchi()
        return self._pseudo_rinchi

    @property
    def pseudo_rinchi_key(self) -> str:
        """Gives a pseudo reaction InChI key"""
        if self._pseudo_rinchi_key is None:
            self._pseudo_rinchi_key = self._make_pseudo_rinchi_key()
        return self._pseudo_rinchi_key

    @property
    def hashed_rid(self) -> str:
        """Gives a reaction hashkey based on Reaction SMILES & reaction id."""
        return hashlib.sha224(f"{self.rsmi}_{self.rid}".encode("utf8")).hexdigest()

    @property
    def reactants_list(self) -> List[str]:
        """Gives all reactants as strings"""
        return split_smiles_from_reaction(self.reactants_smiles)

    @property
    def retro_template(self) -> ReactionTemplate:
        """Gives the retro template"""
        if self._retro_template is None:
            raise ValueError(
                f"Call {self.__class__.__name__}.generate_reaction_template() "
                "before using {self.__class__.__name__}.retro_template"
            )
        return self._retro_template

    @property
    def rinchi(self) -> str:
        """Gives the reaction InChI"""
        if self._rinchi_data is None:
            self._build_rinchi_info()
        return self._rinchi_data.rinchi

    @property
    def rinchi_key_long(self) -> str:
        """Gives the long reaction InChI key"""
        if self._rinchi_data is None:
            self._build_rinchi_info()
        return self._rinchi_data.long_rinchikey

    @property
    def rinchi_key_short(self) -> str:
        """Gives the short reaction InChI key"""
        if self._rinchi_data is None:
            self._build_rinchi_info()
        return self._rinchi_data.short_rinchikey

    def generate_coreagent(self):
        """Extract un-mapped product atoms as extra ractant fragments"""
        if not self.has_partial_mapping():
            return ""

        extra_reactant_fragment = ""
        for product in self.products:
            # Get unmapped atoms
            unmapped_ids = [atom.GetIdx() for atom in product.GetAtoms() if not atom.HasProp("molAtomMapNumber")]
            if len(unmapped_ids) > extractor.MAXIMUM_NUMBER_UNMAPPED_PRODUCT_ATOMS:
                raise ReactionException(f"Template generation failed: too many unmapped atoms ({len(unmapped_ids)})")
            # Define new atom symbols for fragment with atom maps, generalizing fully
            atom_symbols = [f"[{atom.GetSymbol()}]" for atom in product.GetAtoms()]
            # And bond symbols...
            bond_symbols = ["~" for _ in product.GetBonds()]
            if unmapped_ids:
                extra_reactant_fragment += (
                    AllChem.MolFragmentToSmiles(
                        product,
                        unmapped_ids,
                        allHsExplicit=False,
                        isomericSmiles=extractor.USE_STEREOCHEMISTRY,
                        atomSymbols=atom_symbols,
                        bondSymbols=bond_symbols,
                    )
                    + "."
                )

        if extra_reactant_fragment:
            extra_reactant_fragment = extra_reactant_fragment[:-1]
        # Consolidate repeated fragments (stoichometry)
        extra_reactant_fragment = ".".join(sorted(list(set(extra_reactant_fragment.split(".")))))
        return extra_reactant_fragment

    def generate_reaction_template(
        self,
        radius: int = 1,
        expand_ring: bool = False,
        expand_hetero: bool = False,
    ) -> Tuple[ReactionTemplate, ReactionTemplate]:
        """
        Extracts the forward(canonical) and retro reaction template with the specified radius.

        Uses a modified version of:
            https://github.com/connorcoley/ochem_predict_nn/blob/master/data/generate_reaction_templates.py
            https://github.com/connorcoley/rdchiral/blob/master/templates/template_extractor.py

        :param radius: the radius refers to the number of atoms away from the reaction
                       centre to be extracted (the enivronment) i.e. radius = 1 (default)
                       returns the first neighbours around the reaction centre
        :param expand_ring: if True will include all atoms in the same ring as the reaction centre in the template
        :param expand_hetero: if True will extend the template with all bonded hetero atoms
        :returns: the canonical and retrosynthetic templates
        """
        if not self.sanitization_check():
            raise ReactionException("Template generation failed: sanitation check failed")

        reactants = self.reactants
        products = self.products

        for product in products:
            nunmapped = sum(int(not atom.HasProp("molAtomMapNumber")) for atom in product.GetAtoms())
            if nunmapped > extractor.MAXIMUM_NUMBER_UNMAPPED_PRODUCT_ATOMS:
                raise ReactionException(f"Template generation failed: too many unmapped atoms ({nunmapped})")

        if self.no_change():
            raise ReactionException("Template generation failed: no change in reaction")
        if None in reactants + products:
            raise ReactionException("Template generation failed: None in reactants or products")

        extra_atoms = {}
        if expand_ring:
            ring_atom_maps = self._product_unique_ring_atoms()
            for reactant in self.reactants:
                extra_atoms.update(
                    {
                        str(atom.GetAtomMapNum()): atom
                        for atom in reactant.GetAtoms()
                        if atom.GetAtomMapNum() in ring_atom_maps
                    }
                )

        try:
            # Generate canonical reaction template with rdChiral
            canonical_smarts = self._generate_rdchiral_template(
                reactants=reactants,
                products=products,
                radius=radius,
                extra_atoms=extra_atoms,
                expand_hetero=expand_hetero,
            )
            # Canonical ReactionTemplate
            self._canonical_template = ReactionTemplate(canonical_smarts)

            # Retro ReactionTemplate
            reactants_string, _, products_string = split_rsmi(canonical_smarts)
            retro_smarts = products_string + ">>" + reactants_string
            self._retro_template = ReactionTemplate(retro_smarts, "retro")

            # Validate ReactionTemplates
            if not (self._retro_template.rdkit_validation() and self._canonical_template.rdkit_validation()):
                raise ReactionException("Template generation failed: RDkit validation of extracted templates failed")
        except Exception as err:
            raise ReactionException(f"Template generation failed with message: {err}")

        return self.canonical_template, self.retro_template

    @staticmethod
    @wrapt_timeout_decorator.timeout(
        30,
        use_signals=False,
        timeout_exception=ReactionException,
        exception_message="Timed out",
    )
    def _generate_rdchiral_template(
        reactants: List[Chem.Mol],
        products: List[Chem.Mol],
        radius: int = 1,
        extra_atoms: Dict[str, Any] = None,
        expand_hetero: bool = False,
    ) -> str:
        """Generate a reaction template with rdChiral.

        :param reactants: List of reactant molecules
        :param products: List of product molecules
        :param radius: Template radius, defaults to 1
        :raises ReactionException: Template generation failed: could not obtained changed atoms
        :raises ReactionException: Template generation failed: no atoms changes
        :raises ReactionException: Template generation failed: Timed out
        :return: Canonical reaction template
        """
        changed_atoms, changed_atom_tags, err = extractor.get_changed_atoms(reactants=reactants, products=products)

        if expand_hetero:
            for atom in changed_atoms:
                for atom2 in atom.GetNeighbors():
                    if atom2.GetSymbol() not in ["C", "H"] and atom2.GetAtomMapNum():
                        extra_atoms[str(atom2.GetAtomMapNum())] = atom2

        old_tags = list(changed_atom_tags)
        for atom_num, atom in extra_atoms.items():
            if atom_num not in old_tags:
                changed_atoms.append(atom)
                changed_atom_tags.append(atom_num)

        if err:
            raise ReactionException("Template generation failed: could not obtained changed atoms")
        if not changed_atom_tags:
            raise ReactionException("Template generation failed: no atoms changes")

        # Get fragments for reactants
        (
            reactant_fragments,
            _,
            _,
        ) = extractor.get_fragments_for_changed_atoms(
            reactants,
            changed_atom_tags,
            radius=radius,
            expansion=[],
            category="reactants",
        )
        # Get fragments for products %%!
        # (WITHOUT matching groups but WITH the addition of reactant fragments)
        product_fragments, _, _ = extractor.get_fragments_for_changed_atoms(
            products,
            changed_atom_tags,
            radius=radius,
            expansion=extractor.expand_changed_atom_tags(
                changed_atom_tags,
                reactant_fragments,
            ),
            category="products",
        )

        rxn_string = f"{reactant_fragments}>>{product_fragments}"
        canonical_template = extractor.canonicalize_transform(rxn_string)
        # Change from inter-molecular to intra-molecular
        canonical_template_split = split_rsmi(canonical_template)
        canonical_smarts = (
            canonical_template_split[0][1:-1].replace(").(", ".")
            + ">>"
            + canonical_template_split[-1][1:-1].replace(").(", ".")
        )

        return canonical_smarts

    def has_partial_mapping(self) -> bool:
        """Check product atom mapping."""
        for product in self.products:
            sum_with = sum(atom.HasProp("molAtomMapNumber") for atom in product.GetAtoms())
            if sum_with < product.GetNumAtoms():
                return True
        return False

    def is_complete(self) -> bool:
        """Check that the product is not among the reactants"""
        return len(self.reactants) > 0 and len(self.products) > 0

    def no_change(self) -> bool:
        """
        Checks to see if the product appears in the reactant set.

        Compares InChIs to rule out possible variations in SMILES notation.

        :return: True the product is present in the reactants set, else False
        """
        reactant_inchi = {Chem.MolToInchi(reactant) for reactant in self.reactants}
        product_inchi = {Chem.MolToInchi(product) for product in self.products}
        return bool(reactant_inchi & product_inchi)

    def is_fuzzy(self) -> bool:
        """Checks to see if there is fuzziness in the reaction.

        :return: True if there is fuzziness, False otherwise
        """
        return bool(self.rsmi.count("*"))

    def sanitization_check(self) -> bool:
        """
        Checks if the reactant and product mol objects can be sanitized in RDKit.

        The actualy sanitization is carried out when the reaction is instansiated,
        this method will only check that all molecules objects were created.

        :return: True if all the molecule objects were successfully created, else False
        """
        return all(mol is not None for mol in self.reactants + self.agents + self.products)

    def canonical_template_generate_outcome(self) -> bool:
        """Checks whether the canonical template produces"""
        try:
            return len(self.canonical_template.apply(self.reactants_smiles)) > 0
        except Exception:  # pylint: disable=broad-except # noqa
            return False

    def retro_template_generate_outcome(self) -> bool:
        """Checks whether the retrosynthetic template produces an outcome"""
        try:
            return len(self.retro_template.apply(self.products_smiles)) > 0
        except Exception:  # pylint: disable=broad-except # noqa
            return False

    def retro_template_selectivity(self) -> float:
        """
        Checks whether the recorded reactants belong to the set of generated precursors.

        :returns: selectivity, i.e. the fraction of generated precursors matching the recorded precursors
                i.e. 1.0 - match or match.match or match.match.match etc.
                    0.5 - match.none or match.none.match.none etc.
                    0.0 - none
        """
        # pylint: disable=too-many-branches
        try:
            retro_outcomes = self.retro_template.apply(self.products_smiles)
        except Exception:  # pylint: disable=broad-except # noqa
            return 0.0

        if not extractor.USE_STEREOCHEMISTRY:
            reactant_mol_list = []
            for smiles in self.reactants_list:
                if not smiles:
                    continue
                reactant_mol_list.append(Chem.MolFromSmiles(smiles.replace("@", "")))
        else:
            reactant_mol_list = self.reactants

        reactant_inchi = [Chem.MolToInchi(reactant) for reactant in reactant_mol_list]
        precursor_set = []
        for outcome_set in retro_outcomes:
            if not extractor.USE_STEREOCHEMISTRY:
                outcome_set_inchi = [
                    Chem.MolToInchi(Chem.MolFromSmiles(outcome.replace("@", ""))) for outcome in outcome_set
                ]
            else:
                outcome_set_inchi = [Chem.MolToInchi(Chem.MolFromSmiles(outcome)) for outcome in outcome_set]
            precursor_set.append(outcome_set_inchi)

        assessment = []
        for precursor in precursor_set:
            # There must be a match between the generated outcomes and recorded reactants
            if len(list(set(precursor) & set(reactant_inchi))) != 0:
                assessment.append(2)
            # No match or error
            elif len(list(set(precursor) & set(reactant_inchi))) == 0:
                assessment.append(1)
            else:
                print("Template error")
                assessment.append(0)

        # Quantify the level of selectivity, if an error has occured set to 0
        if assessment.count(0) != 0:
            return 0
        return assessment.count(2) / len(assessment)

    def _build_rinchi_info(self) -> None:
        try:
            rinchi_data = rinchi_api.generate_rinchi(self.rsmi)
        except rinchi_api.RInChIError as err:
            raise ReactionException(f"Could not generate reaction InChI: {err}")
        self._rinchi_data = rinchi_data

    def _clean_reaction(self) -> None:
        """This function updates the reaction smiles by moving mapped molecules from agents to reactants and
        unmapped molecules from reactants to the agents to have reaction smiles in the form of:
            mapped_reactants>unmapped_reagents>products

        Stores:
            rsmi (str): Reaction SMILES in the form mapped_reactants>unmapped_reagents>products
            clean_rsmi (str): Reaction SMILES in the form mapped_reactants>unmapped_reagents>mapped_products
        """

        def _atom_mapping(smiles: str) -> Tuple[int, int]:
            """Return the number of atoms and number of mapped atoms of molecule.

            :param smiles: SMILES of molecule
            :type smiles: string
            :return: Numbers of atoms and mapped atoms of molecule.
            :rtype: Tuple[int, int]
            """
            mol = Chem.MolFromSmiles(smiles)
            num_atoms_ = 0
            num_mapped_atoms_ = 0
            if mol:
                num_mapped_atoms_ = sum(atom.HasProp("molAtomMapNumber") for atom in mol.GetAtoms())
                num_atoms_ = mol.GetNumAtoms()

            return num_atoms_, num_mapped_atoms_

        split_reaction = split_rsmi(self.rsmi)
        self.reactants_smiles = extractor.replace_deuterated(split_reaction[0])
        self.agents_smiles = extractor.replace_deuterated(split_reaction[1])
        self.products_smiles = extractor.replace_deuterated(split_reaction[2])

        # If any agent has mapped atoms then move it to reactants
        updated_agents = []
        add_reactants = []
        for agent in self.agents_list:
            _, num_mapped_atoms = _atom_mapping(agent)
            if num_mapped_atoms:
                add_reactants.append(agent)
            else:
                updated_agents.append(agent)
        self.agents_smiles = self._join_components(updated_agents)

        # If any reactant has 0 mapped atoms then move it to agents
        updated_reactants = []
        add_agents = []
        for reactant in self.reactants_list:
            _, num_mapped_atoms = _atom_mapping(reactant)
            if num_mapped_atoms == 0:
                add_agents.append(reactant)
            else:
                updated_reactants.append(reactant)
        self.reactants_smiles = self._join_components(updated_reactants)

        # Update reactants with molecules from agents
        if add_reactants:
            self.reactants_smiles = ".".join([self.reactants_smiles, self._join_components(add_reactants)])
        # Update agents with molecules from reactants
        if add_agents:
            if self.agents_smiles:
                self.agents_smiles = ".".join([self.agents_smiles, self._join_components(add_agents)])
            else:
                self.agents_smiles = self._join_components(add_agents)

        # Neutralize/Uncharge Reactants & Products
        # Neutralize reactants
        neutral_reactants = utils.neutralize_molecules(self.reactants_list)
        self.reactants_smiles = self._join_components(neutral_reactants)
        # Neutralize products
        neutral_products = utils.neutralize_molecules(self.products_list)
        self.products_smiles = self._join_components(neutral_products)

        # Reaction SMILES
        self.rsmi = ">".join([self.reactants_smiles, self.agents_smiles, self.products_smiles])

        # Clean Reaction SMILES by RDKit
        reaction = AllChem.ReactionFromSmarts(self.rsmi)
        reaction.Initialize()
        reaction.RemoveUnmappedReactantTemplates(thresholdUnmappedAtoms=0.01)
        reaction.RemoveUnmappedProductTemplates(thresholdUnmappedAtoms=0.01)
        self.clean_rsmi = AllChem.ReactionToSmiles(reaction)

    def _join_components(self, list_: List[str]) -> str:
        """Join SMILES components to one SMILES.

        :param list_: List of SMILES components
        :return: A single SMILES
        """
        return ".".join([f"({item})" if "." in item else item for item in list_])

    def _product_unique_ring_atoms(self) -> List[int]:
        def find_ring_atom_maps(mol):
            ring_atoms_sets = []
            for ring in mol.GetRingInfo().AtomRings():
                ring_atoms_sets.append(tuple(sorted(mol.GetAtomWithIdx(idx).GetAtomMapNum() for idx in ring)))
            return ring_atoms_sets

        reactant_ring_atom_sets = []
        for reactant in self.reactants:
            reactant_ring_atom_sets.extend(find_ring_atom_maps(reactant))

        product_ring_atom_sets = []
        for product in self.products:
            for set_ in find_ring_atom_maps(product):
                if set_ not in reactant_ring_atom_sets:
                    product_ring_atom_sets.append(set_)

        return [idx for ring_atoms in product_ring_atom_sets for idx in ring_atoms]

    def _split_reaction(self) -> None:
        """
        This function is used instead of `_clean_reaction` if cleaning is not required.

        Stores:
            rsmi (str): Reaction SMILES in the form mapped_reactants>unmapped_reagents>products
        """
        split_reaction = split_rsmi(self.rsmi)
        self.reactants_smiles = extractor.replace_deuterated(split_reaction[0])
        self.agents_smiles = extractor.replace_deuterated(split_reaction[1])
        self.products_smiles = extractor.replace_deuterated(split_reaction[2])
        self.rsmi = ">".join([self.reactants_smiles, self.agents_smiles, self.products_smiles])
        self.clean_rsmi = self.rsmi

    def _make_pseudo_rinchi(self) -> str:
        """Generate pseudo Reaction InChI by concatenate the InChIs of reactants and products.

        :return: Pseudo Reaction InChI
        :rtype: str
        """
        reactant_inchi = (
            Chem.MolToInchi(Chem.MolFromSmiles(utils.remove_atom_mapping(self.reactants_smiles)))
            if self.reactants_smiles
            else ""
        )
        product_inchi = (
            Chem.MolToInchi(Chem.MolFromSmiles(utils.remove_atom_mapping(self.products_smiles)))
            if self.products_smiles
            else ""
        )
        concatenated_rinchi = "++".join([reactant_inchi, product_inchi])
        return concatenated_rinchi

    def _make_pseudo_rinchi_key(self) -> str:
        """Generate pseudo Reaction InChI-Key by concatenate the InChIs of reactants and products.

        :return: Pseudo Reaction InChI-Key
        :rtype: str
        """
        reactant_inchi_key = Chem.MolToInchiKey(Chem.MolFromSmiles(utils.remove_atom_mapping(self.reactants_smiles)))
        product_inchi_key = Chem.MolToInchiKey(Chem.MolFromSmiles(utils.remove_atom_mapping(self.products_smiles)))
        concatenated_rinchi_key = "++".join([reactant_inchi_key, product_inchi_key])
        return concatenated_rinchi_key



""" Module containing the implementation of the SC-score model for synthetic complexity scoring. """

from __future__ import annotations

import numpy as np
from rdkit.Chem import AllChem

RdMol = AllChem.rdchem.Mol

try:
    import onnxruntime
except ImportError:
    SCORE_SUPPORTED = False
else:
    SCORE_SUPPORTED = True


class SCScore:
    """
    Encapsulation of the SCScore model

    Re-write of the SCScorer from the scscorer package using ONNX

    The predictions of the score is made with a sanitized instance of an RDKit molecule

    .. code-block::

        mol = Chem.MolFromSmiles(CCC)
        scscorer = SCScorer("path_to_model")
        score = scscorer(mol)

    :param model_path: the filename of the model weights and biases
    :param fingerprint_length: the number of bits in the fingerprint
    :param fingerprint_radius: the radius of the fingerprint
    """

    def __init__(
        self,
        model_path: str,
        fingerprint_length: int = 1024,
        fingerprint_radius: int = 2,
    ) -> None:
        if not SCORE_SUPPORTED:
            raise ImportError("Cannot use SCScore model. Install rxnutils with optional dependencies")
        self._fingerprint_length = fingerprint_length
        self._fingerprint_radius = fingerprint_radius
        session_options = onnxruntime.SessionOptions()
        session_options.intra_op_num_threads = 1
        self._model = onnxruntime.InferenceSession(model_path, sess_options=session_options)

    def __call__(self, rd_mol: RdMol) -> float:
        fingerprint = self._make_fingerprint(rd_mol)
        return self._model.run(None, {"fingerprint": fingerprint})[0][0]

    def _make_fingerprint(self, rd_mol: RdMol) -> np.ndarray:
        """Returns the molecule's Morgan fingerprint"""
        fp_vec = AllChem.GetMorganFingerprintAsBitVect(
            rd_mol,
            self._fingerprint_radius,
            nBits=self._fingerprint_length,
            useChirality=True,
        )
        return np.array(fp_vec, dtype=float)
